#! /usr/bin/env python3

import mpqp.execution.connection.aws_connection as awsc
import mpqp.execution.connection.env_manager as env_m
import mpqp.execution.connection.ibm_connection as ibmqc
import mpqp.execution.connection.qlm_connection as qlmc
from mpqp.tools.choice_tree import AnswerNode, QuestionNode, run_choice_tree
from mpqp.tools.errors import IBMRemoteExecutionError


def print_config_info():
    """
    Prints the info concerning each provider's registered account
    """
    print("===== IBM Quantum info : ===== ")
    try:
        print(ibmqc.get_active_account_info())
    except IBMRemoteExecutionError as err:
        if "No IBM Q account configured" in str(err):
            print("Account not configured")

    print("===== Qaptiva QLMaaS info : ===== ")
    user_name = env_m.get_env_variable("QLM_USER")
    if user_name == "":
        print("Account not configured")
    else:
        print("Current user: ", env_m.get_env_variable("QLM_USER"))
    print("===== AWS Braket info : ===== ")
    try:
        print(awsc.get_aws_braket_account_info())
    except Exception as err:
        if "No AWS Braket account configured" in str(err):
            print("Account not configured")
        else:
            print("Error occurred when getting AWS account info.")
    input("Press 'Enter' to continue")
    return "", []


def main_setup():
    """Function called by the script bellow to setup all connections or to get
    information about the existing ones."""

    setup_tree = QuestionNode(
        "~~~~~ MPQP REMOTE CONFIGURATION ~~~~~",
        [
            AnswerNode("IBM configuration", ibmqc.setup_ibm_account),
            AnswerNode("QLM configuration", qlmc.setup_qlm_account),
            AnswerNode("Amazon Braket configuration", awsc.setup_aws_braket_account),
            AnswerNode("Config information", print_config_info),
        ],
    )

    for answer in setup_tree.answers:
        answer.next_question = setup_tree

    run_choice_tree(setup_tree)


if __name__ == "__main__":
    main_setup()



from getpass import getpass

from qiskit.providers import QiskitBackendNotFoundError
from qiskit.providers.backend import BackendV1
from qiskit_ibm_provider import IBMProvider
from qiskit_ibm_provider.accounts import AccountNotFoundError
from qiskit_ibm_provider.api.exceptions import RequestsApiError
from qiskit_ibm_runtime import QiskitRuntimeService
from termcolor import colored
from typeguard import typechecked

from mpqp.execution.connection.env_manager import get_env_variable, save_env_variable
from mpqp.execution.devices import IBMDevice
from mpqp.tools.errors import IBMRemoteExecutionError

Ibm_Provider = None
Runtime_Service = None


@typechecked
def config_ibm_account(token: str):
    """
    Configure and save locally IBM Quantum account's information.

    Args:
        token: IBM Quantum API token.

    Raises:
        IBMRemoteExecutionError
    """

    try:
        IBMProvider.save_account(token=token, overwrite=True)
        save_env_variable("IBM_CONFIGURED", "True")
        save_env_variable("IBM_TOKEN", token)
    except Exception as err:
        # if an error occurred, we put False in the mpqp config file
        save_env_variable("IBM_CONFIGURED", "False")
        raise IBMRemoteExecutionError(
            "Error when saving the account.\nTrace: " + str(err)
        )


def setup_ibm_account():
    """Setups and updates the IBM Q account using the existing configuration and by asking for the token ."""
    was_configured = get_env_variable("IBM_CONFIGURED") == "True"

    if was_configured:
        decision = input(
            "An IBMQ account is already configured. Do you want to update it? [y/N]"
        )
        if decision.lower().strip() != "y":
            return "Canceled.", []

    token = getpass("Enter your IBM token (hidden): ")
    if token == "":
        print(colored("Empty credentials", "red"))
        getpass("Press 'Enter' to continue")
        return "", []
    old_token = get_env_variable("IBM_TOKEN")
    config_ibm_account(token)
    if test_connection():
        return "IBM Q account correctly configured", []
    else:
        if was_configured:
            config_ibm_account(old_token)
        else:
            save_env_variable("IBM_CONFIGURED", "False")
        getpass("Press 'Enter' to continue")
        return "", []


def test_connection() -> bool:
    """Tests if the connection to the provider works.

    Returns:
        ``False`` if login failed.
    """
    try:
        IBMProvider()
    except RequestsApiError as err:
        if "Login failed" in str(err):
            print(colored("Wrong credentials", "red"))
            return False
        else:
            raise err
    return True


def get_IBMProvider() -> IBMProvider:
    """Returns the IBMProvider needed to get one or several backends for
    execution.

    Example:
        >>> instance = get_IBMProvider()
        >>> instance.backends()
        [<IBMBackend('ibmq_qasm_simulator')>,
         <IBMBackend('simulator_extended_stabilizer')>,
         <IBMBackend('simulator_mps')>,
         <IBMBackend('simulator_stabilizer')>,
         <IBMBackend('simulator_statevector')>,
         <IBMBackend('ibm_brisbane')>,
         <IBMBackend('ibm_kyoto')>,
         <IBMBackend('ibm_osaka')>]

    Raises:
        IBMRemoteExecutionError
    """
    global Ibm_Provider
    if Ibm_Provider is None:
        if get_env_variable("IBM_CONFIGURED") == "False":
            raise IBMRemoteExecutionError(
                "Error when instantiating IBM Provider. No IBM Q account configured."
            )
        try:
            Ibm_Provider = IBMProvider()
        except RequestsApiError as err:
            raise IBMRemoteExecutionError(
                "Error when instantiating IBM Provider (probably wrong token saved "
                "in the account).\nTrace: " + str(err)
            )
        except AccountNotFoundError as err:
            raise IBMRemoteExecutionError(
                "Error when instantiating IBM Provider. No IBM Q account configured.\nTrace: "
                + str(err)
            )
    return Ibm_Provider


def get_QiskitRuntimeService() -> QiskitRuntimeService:
    """
    Returns the QiskitRuntimeService needed for remote connection and execution

    Example:
        >>> service = get_QiskitRuntimeService()
        >>> service.jobs()
        [<RuntimeJob('cmdj3b4nktricigarn8g', 'estimator')>,
         <RuntimeJob('cmdj3a74mi97k7j7ujv0', 'sampler')>,
         <RuntimeJob('cmama29054sir2cq94og', 'estimator')>,
         <RuntimeJob('cmama14pduldih1q4ktg', 'sampler')>,
         <RuntimeJob('cm7vds4pduldih1k1mq0', 'sampler')>]

    Raises:
        IBMRemoteExecutionError
    """
    global Runtime_Service
    if Runtime_Service is None:
        if get_env_variable("IBM_CONFIGURED") == "False":
            raise IBMRemoteExecutionError(
                "Error when instantiating QiskitRuntimeService. No IBM account configured."
            )
        try:
            Runtime_Service = QiskitRuntimeService()
        except Exception as err:
            raise IBMRemoteExecutionError(
                "Error when instantiating QiskitRuntimeService (probably wrong token saved "
                "in the account).\nTrace: " + str(err)
            )
    return Runtime_Service


def get_active_account_info() -> str:
    """
    Returns the information concerning the active IBMQ account

    Example:
        >>> print(get_active_account_info())
            Channel: ibm_quantum
            Instance: ibm-q-startup/colibritd/default
            Token: bf5e5*****
            URL: https://auth.quantum-computing.ibm.com/api
            Verify: True

    Returns:
        A string describing the account info.
    """
    provider = get_IBMProvider()
    account = provider.active_account()
    assert account is not None
    return f"""    Channel: {account["channel"]}
    Instance: {account["instance"]}
    Token: {account["token"][:5]}*****
    URL: {account["url"]}
    Verify: {account["verify"]}"""


@typechecked
def get_backend(device: IBMDevice) -> BackendV1:
    """
    Retrieves the IBM Q remote device corresponding to the device in parameter

    Args:
        device: The IBMDevice to get from IBMQ provider.

    Example:
        >>> brisbane = get_backend(IBMDevice.IBM_BRISBANE)
        >>> brisbane.properties().gates[0].parameters
        [Nduv(datetime.datetime(2024, 1, 9, 11, 3, 18, tzinfo=tzlocal()), gate_error, , 0.00045619997922344296),
         Nduv(datetime.datetime(2024, 1, 9, 15, 41, 39, tzinfo=tzlocal()), gate_length, ns, 60)]

    Returns:
        A qiskit.providers.backend.Backend object that will be use to execute circuit.

    Raises:
        IBMRemoteExecutionError
    """
    # NOTE:
    #       Question : when a backend is present in several IBMQ instances, which instance does it use to submit jobs
    # on this backend ? Typically if with colibritd instance i have more priority and by default it uses ibmq
    # instance, then i lose something here.
    #       Answer : it takes the default instance attached to the account (the higher plan, usually).
    if not device.is_remote():
        raise ValueError("Expected a remote IBMQ device but got a local simulator.")

    provider = get_IBMProvider()

    try:
        backend = provider.get_backend(device.value)

    except QiskitBackendNotFoundError as err:
        raise IBMRemoteExecutionError(
            f"Requested device {device} not found. Verify if your instances "
            "allows to access this machine, or the device's name.\n"
            f"Trace: {err}"
        )

    return backend


def get_all_job_ids() -> list[str]:
    """
    Retrieves all the job ids of this account from the several IBM remote providers
    (IBMProvider, QiskitRuntimeService, ...)

    Example:
        >>> get_all_job_ids()
        ['cm6pp7e879ps6bbo7m30', 'cm6ou0q70abqioeudkd0', 'cm6opgcpduldih1hq7j0', 'cm01vp4pduldih0uoi2g',
        'cnvw8z3b08x0008y3e4g', 'cnvw7qyb08x0008y3e0g', 'cnvw7fdvn4c0008a6ztg', 'cnvw79dvn4c0008a6zt0',
        'cnvw64rb08x0008y3dx0', 'cnvw5z7wsx00008wybcg', 'cmdj3b4nktricigarn8g', 'cmdj3a74mi97k7j7ujv0',
        'cmama29054sir2cq94og', 'cmama14pduldih1q4ktg', 'cm80qmi70abqiof0o170', 'cm80qlkpduldih1k4png',
        'cm80pb1054sir2ck9i3g', 'cm80pa6879ps6bbqg2pg', 'cm7vdugiidfp3m8rg02g', 'cm7vds4pduldih1k1mq0']
    """
    all_job_ids = []

    ibm_provider = get_IBMProvider()  # using IBMProvider
    service = get_QiskitRuntimeService()  # using QiskitRuntimeService

    if ibm_provider:
        all_job_ids.extend([job.job_id() for job in ibm_provider.jobs()])

    if service:
        all_job_ids.extend([job.job_id() for job in service.jobs()])

    return all_job_ids



from mpqp.gates import *
from mpqp import QCircuit
from mpqp.measures import BasisMeasure

# Declaration of the circuit with the right size
circuit = QCircuit(4)

# Constructing the circuit by adding gates
circuit.add(T(0))
circuit.add(CNOT(0, 1))
circuit.add(X(0))
circuit.add(H(1))
circuit.add(Z(2))
circuit.add(CZ(2, 1))
circuit.add(SWAP(2, 0))
circuit.add(CNOT(0, 2))
circuit.add(Ry(3.14 / 2, 2))
circuit.add(S(1))
circuit.add(H(3))
circuit.add(CNOT(1, 2))
circuit.add(Rx(3.14, 1))
circuit.add(CNOT(3, 0))
circuit.add(Rz(3.14, 0))
# Add measurement
circuit.add(BasisMeasure([0, 1, 2, 3], shots=2000))

print(circuit.to_qasm2())
print(circuit.to_qasm3())


# Declaration of the circuit with the right size
circuit = QCircuit(2)

# Constructing the circuit by adding gates
circuit.add(X(0))
circuit.add(CNOT(0, 1))
# Add measurement
circuit.add(BasisMeasure([1], shots=2000))

print(circuit.to_qasm2())
print(circuit.to_qasm3())



from dataclasses import dataclass
from typing import Any, Callable, Iterable, Optional, TypeVar
from pick import pick
from typeguard import typechecked

from mpqp.tools.generics import find

T = TypeVar("T")


@dataclass
class AnswerNode:
    """Represents a node in a decision tree corresponding to an answer to a question. An answer can lead to an action,
    or to another question.

    Args:
        label: See attribute description.
        action: See attribute description.
        next_question: See attribute description.
    """

    label: str
    """The label or text associated with the answer."""
    action: Callable[..., tuple[str, Iterable[Any]]]
    """A callable representing an action function linked to an answer."""
    next_question: Optional["QuestionNode"] = None
    """An optional reference to the next question node."""


@dataclass
class QuestionNode:
    """Represents a node in a decision tree corresponding to a question.

    Args:
        label: See attribute description.
        answers: See attribute description.
    """
    label: str
    """The label or text associated with the question."""
    answers: list[AnswerNode]
    """List of possible answers to this question."""


@typechecked
def run_choice_tree(question: QuestionNode):
    """
    Execute the choice tree by starting with the question node in parameter.

    Args:
        question: Root question from which the choice tree will start
    """
    # 3M-TODO: allow exit with Ctrl+C and Q
    prev_args = []
    prev_message = ""
    next_question = question
    while True:
        option, _ = pick(
            list(map(lambda a: a.label, next_question.answers)) + ["Exit"],
            prev_message + "\n\n" + next_question.label,
            indicator="=>",
        )
        if option == "Exit":
            return
        selected_answer = find(next_question.answers, lambda a: a.label == option)
        prev_message, prev_args = selected_answer.action(*prev_args)
        next_question = selected_answer.next_question
        if next_question is None:
            pick(["Press 'Enter' to continue"], prev_message, indicator="")
            return


# Example:
if __name__ == "__main__":

    def date_name_picker():
        name = input("What's your name?\n\t")
        return f"Then it's a date {name}!", [name]

    choice_tree = QuestionNode(
        "Are you free?",
        [
            AnswerNode(
                "Yes",
                date_name_picker,
                QuestionNode(
                    "Do you have a preference for the day?",
                    [
                        AnswerNode(
                            "Not at all", lambda n: (f"Tomorrow then {n} :D", [])
                        ),
                        AnswerNode("Tomorrow", lambda n: (f"Tomorrow then {n} :D", [])),
                    ],
                ),
            ),
            AnswerNode(
                "No",
                lambda: ("What a shame :(", []),
                QuestionNode(
                    "Would you be willing to reconsider ?",
                    [
                        AnswerNode("Yes", lambda: ("Cool!", [])),
                        AnswerNode("No", lambda: ("Sad :'(", [])),
                    ],
                ),
            ),
        ],
    )

    assert choice_tree.answers[-1].next_question is not None

    choice_tree.answers[-1].next_question.answers[0].next_question = choice_tree

    run_choice_tree(choice_tree)



from __future__ import annotations

import math
from typing import Optional

from qiskit import QuantumCircuit
from qiskit_aer import Aer
from qiskit_aer import AerSimulator
from qiskit.providers import BackendV1, BackendV2, JobStatus as IBM_JobStatus
from qiskit.compiler import transpile
from qiskit.primitives import (
    Estimator as Qiskit_Estimator,
    BackendEstimator,
    EstimatorResult,
    SamplerResult,
)
from qiskit.quantum_info import Operator
from qiskit.result import Result as QiskitResult
from qiskit_ibm_provider.job import IBMJob
from qiskit_ibm_runtime import (
    RuntimeJob,
    Sampler as Runtime_Sampler,
    Estimator as Runtime_Estimator,
    Session,
)
from typeguard import typechecked

from mpqp.core.circuit import QCircuit
from mpqp.core.instruction.measurement import BasisMeasure
from mpqp.core.instruction.measurement.expectation_value import ExpectationMeasure
from mpqp.execution.connection.ibm_connection import (
    get_IBMProvider,
    get_QiskitRuntimeService,
)
from mpqp.execution.devices import IBMDevice
from mpqp.execution.job import Job, JobType, JobStatus
from mpqp.execution.result import StateVector, Sample, Result
from mpqp.core.languages import Language
from mpqp.tools.errors import DeviceJobIncompatibleError, IBMRemoteExecutionError


@typechecked
def run_ibm(job: Job) -> Result:
    """
    Execute the job on the right IBM Q device precised in the job in parameter.
    This function is not meant to be used directly, please use ``runner.run(...)`` instead.

    Args:
        job: Job to be executed

    Returns:
        a Result after submission and execution of the job
    """
    return run_aer(job) if not job.device.is_remote() else run_ibmq(job)


@typechecked
def compute_expectation_value(
    ibm_circuit: QuantumCircuit, ibm_backend: Optional[BackendV1 | BackendV2], job: Job
) -> Result:
    """
    Configures observable job and run it locally, and returns the corresponding Result.
    This function is not meant to be used directly, please use ``runner.run(...)`` instead.

    Args:
        ibm_circuit: QuantumCircuit (already reversed bits)
        ibm_backend: The IBM backend (local of remote) on which we execute the job.
        job: Mpqp job describing the observable job to run.

    """
    if not isinstance(job.measure, ExpectationMeasure):
        raise ValueError(
            "Cannot compute expectation value if measure used in job is not of "
            "type ExpectationMeasure"
        )
    nb_shots = job.measure.shots
    qiskit_observable = Operator(job.measure.observable.matrix)

    if nb_shots != 0:
        assert ibm_backend is not None
        estimator = BackendEstimator(backend=ibm_backend)
    else:
        estimator = Qiskit_Estimator()

    # 6M-TODO : think of the possibility to compute several expectation values at the same time when the circuit is
    #  the same apparently the estimator.run() can take several circuits and observables at the same time,
    #  to verify if putting them all together increases the performance

    job.status = JobStatus.RUNNING
    job_expectation = estimator.run(
        [ibm_circuit], [qiskit_observable], shots=nb_shots if nb_shots != 0 else None
    )
    estimator_result = job_expectation.result()
    assert isinstance(job.device, IBMDevice)
    return extract_result(estimator_result, job, job.device)


@typechecked
def check_job_compatibility(job: Job):
    """
    Checks whether the job in parameter has coherent and compatible attributes.

    Args:
        job: Job for which we want to check compatibility.

    Raises:
        DeviceJobIncompatibleError
    """
    if not type(job.measure) in job.job_type.value:
        raise DeviceJobIncompatibleError(
            f"An {job.job_type.name} job is valid only if the corresponding circuit has an measure in "
            f"{list(map(lambda cls: cls.__name__, job.job_type.value))}. "
            f"{type(job.measure).__name__} was given instead."
        )
    if (
        job.job_type == JobType.STATE_VECTOR
        and job.device != IBMDevice.AER_SIMULATOR_STATEVECTOR
    ):
        raise DeviceJobIncompatibleError(
            "Cannot reconstruct state vector with this device. Please use "
            f"{IBMDevice.AER_SIMULATOR_STATEVECTOR} instead"
        )
    if job.device == IBMDevice.AER_SIMULATOR_STATEVECTOR:
        if job.job_type == JobType.SAMPLE:
            raise DeviceJobIncompatibleError(
                "Cannot use sample mode with the statevector simulator."
            )
        if job.job_type == JobType.OBSERVABLE:
            assert job.measure is not None
            if job.measure.shots > 0:
                raise DeviceJobIncompatibleError(
                    "Cannot compute expectation values with non-zero shots"
                    f" with {IBMDevice.AER_SIMULATOR_STATEVECTOR}.\nSet the"
                    " shots to zero to get the exact value, or select "
                    "another device instead"
                )


@typechecked
def run_aer(job: Job):
    """
    Executes the job on the right AER local simulator precised in the job in parameter.
    This function is not meant to be used directly, please use ``runner.run(...)`` instead.

    Args:
        job: Job to be executed.

    Returns:
        A Result after submission and execution of the job.
    """
    qiskit_circuit = (
        job.circuit.without_measurements().to_other_language(Language.QISKIT)
        if (job.job_type == JobType.STATE_VECTOR)
        else job.circuit.to_other_language(Language.QISKIT)
    )

    qiskit_circuit = qiskit_circuit.reverse_bits()
    check_job_compatibility(job)

    # define backend simulator
    if job.device == IBMDevice.AER_SIMULATOR:
        backend_sim = AerSimulator()
        if job.job_type == JobType.SAMPLE:
            assert job.measure is not None
            run_input = transpile(qiskit_circuit, backend_sim)
            job.status = JobStatus.RUNNING
            job_sim = backend_sim.run(run_input, shots=job.measure.shots)
            result_sim = job_sim.result()
            result = extract_result(result_sim, job, IBMDevice.AER_SIMULATOR)
        elif job.job_type == JobType.OBSERVABLE:
            result = compute_expectation_value(qiskit_circuit, backend_sim, job)
        else:
            raise ValueError(f"Job type {job.job_type} not handled on {job.device}")

    elif job.device == IBMDevice.AER_SIMULATOR_STATEVECTOR:
        if job.job_type == JobType.STATE_VECTOR:
            backend_sim = Aer.get_backend(job.device.value)
            # the save_statevector method is patched on qiskit_aer load, meaning
            # the type checker can't find it. I hate it but it is what it is.
            # this explains the `type: ignore`. This method is needed to get a
            # statevector our of the statevector simulator...
            qiskit_circuit.save_statevector()  # type: ignore
            job.status = JobStatus.RUNNING
            job_sim = backend_sim.run(qiskit_circuit, shots=0)
            result_sim = job_sim.result()
            result = extract_result(
                result_sim, job, IBMDevice.AER_SIMULATOR_STATEVECTOR
            )
        elif job.job_type == JobType.OBSERVABLE:
            result = compute_expectation_value(qiskit_circuit, None, job)
        else:
            raise ValueError(f"job type {job.job_type} not handled on {job.device}")

    else:
        raise ValueError(f"job device {job.device} not handled yet")

    job.status = JobStatus.DONE
    return result


@typechecked
def submit_ibmq(job: Job) -> tuple[str, RuntimeJob | IBMJob]:
    """
    Submits the job on the remote IBM device (quantum computer or simulator).
    This function is not meant to be used directly, please use ``runner.submit(...)`` instead.

    Args:
        job: Job to be executed.

    Returns:
        The job_id of the submitted job and the RuntimeJob/IBMCircuitJob.
    """
    if job.job_type == JobType.STATE_VECTOR:
        raise DeviceJobIncompatibleError(
            "State vector cannot be computed using IBM remote simulators and"
            " devices. Please use a local simulator instead."
        )

    if job.job_type == JobType.OBSERVABLE:
        if not isinstance(job.measure, ExpectationMeasure):
            raise ValueError(
                "An observable job must is valid only if the corresponding "
                "circuit has an expectation measure."
            )
        if job.measure.shots == 0:
            raise DeviceJobIncompatibleError(
                "Expectation values cannot be computed exactly using IBM remote"
                " simulators and devices. Please use a local simulator instead."
            )
    check_job_compatibility(job)

    qiskit_circuit = job.circuit.to_other_language(Language.QISKIT).reverse_bits()

    service = get_QiskitRuntimeService()
    backend_str = job.device.value
    session = Session(service=service, backend=backend_str)
    if job.job_type == JobType.OBSERVABLE:
        assert isinstance(job.measure, ExpectationMeasure)
        estimator = Runtime_Estimator(session=session)
        qiskit_observable = Operator(job.measure.observable.matrix)

        ibm_job = estimator.run(
            qiskit_circuit, qiskit_observable, shots=job.measure.shots
        )
    elif job.job_type == JobType.SAMPLE:
        assert job.measure is not None
        sampler = Runtime_Sampler(session=session)
        ibm_job = sampler.run(qiskit_circuit, shots=job.measure.shots)
    else:
        raise NotImplementedError(f"{job.job_type} not handled.")

    job.id = ibm_job.job_id()

    return job.id, ibm_job


@typechecked
def run_ibmq(job: Job) -> Result:
    """
    Submits the job on the right IBMQ remote device, precised in the job in parameter, and waits until the job is
    completed.
    This function is not meant to be used directly, please use ``runner.run(...)`` instead.

    Args:
        job: Job to be executed.

    Returns:
        A Result after submission and execution of the job.
    """
    _, remote_job = submit_ibmq(job)
    ibm_result = remote_job.result()

    assert isinstance(job.device, IBMDevice)
    return extract_result(ibm_result, job, job.device, remote_job)


@typechecked
def extract_result(
    result: QiskitResult | EstimatorResult | SamplerResult,
    job: Optional[Job] = None,
    device: IBMDevice = IBMDevice.AER_SIMULATOR,
    ibm_job: Optional[IBMJob | RuntimeJob] = None,
) -> Result:
    """
    Parses a result from IBM execution (remote or local) into an mpqp Result.
    Depending on which service you run the job (local/remote backend, Estimator, Sampler),
    you retrieve a different result.

    Args:
        result: Result returned by IBM after running of the job.
        job: Original mpqp job used to generate the run. Used to retrieve more easily info to instantiate the result.
        device: IBMDevice on which the job was submitted. Used to know if the run was remote or local
        ibm_job: IBM or Runtime job used to retrieve info about the circuit and the submitted job (in the remote case).

    Returns:
        A Result containing the result info extracted from the IBM result.
    """

    if job is not None and (
        isinstance(result, EstimatorResult) != (job.job_type == JobType.OBSERVABLE)
    ):
        raise ValueError(
            "Mismatch between job type and result type: either the result is an"
            " `EstimatorResult` and the job is od type of both those assertions"
            " are false."
        )

    if isinstance(result, EstimatorResult):
        if job is None:
            job = Job(JobType.OBSERVABLE, QCircuit(0), device)
        shots = 0 if len(result.metadata[0]) == 0 else result.metadata[0]["shots"]
        variance = (
            None if len(result.metadata[0]) == 0 else result.metadata[0]["variance"]
        )
        return Result(job, result.values[0], variance, shots)

    elif isinstance(result, SamplerResult):
        shots = result.metadata[0]["shots"]
        probas = result.quasi_dists[0]
        if job is None:
            if ibm_job is None:
                #  If we don't have access to the remote Sampler RuntimeJob, we determine the number of qubits by taking
                #  the max index in the counts and take the upper power of two. Of course this is not a clean way and
                #  can lead to a lower number of qubit than the real one. We asked IBM support, apparently there is no
                #  way to retrieve the right nb_qubits with SamplerResult only. That is why we encourage to input the
                #  ibm_job to this function
                max_index = max(list(probas.keys()))
                nb_qubits = math.ceil(math.log2(max_index + 1))
            else:
                if isinstance(ibm_job, RuntimeJob):
                    nb_qubits = len(ibm_job.inputs["circuits"][0].qubits)
                else:
                    raise ValueError(
                        f"Expected a RuntimeJob as optional parameter but got an {type(ibm_job)} instead"
                    )
            job = Job(
                JobType.SAMPLE,
                QCircuit(nb_qubits),
                device,
                BasisMeasure(list(range(nb_qubits)), shots=shots),
            )

        data = [
            Sample(
                index=item, probability=probas[item], nb_qubits=job.circuit.nb_qubits
            )
            for item in probas
        ]
        return Result(job, data, None, shots)

    elif isinstance(
        result, QiskitResult
    ):  # pyright: ignore[reportUnnecessaryIsInstance]
        if job is None:
            job_data = result.data()
            if "statevector" in job_data:
                job_type = JobType.STATE_VECTOR
                nb_qubits = int(math.log(len(result.get_statevector()), 2))
                job = Job(job_type, QCircuit(nb_qubits), device)
            elif "counts" in job_data:
                job_type = JobType.SAMPLE
                nb_qubits = len(list(result.get_counts())[0])
                shots = result.results[0].shots
                job = Job(
                    job_type,
                    QCircuit(nb_qubits),
                    device,
                    BasisMeasure(list(range(nb_qubits)), shots=shots),
                )
            else:
                if len(result.data()) == 0:
                    raise ValueError(
                        "Result data is empty, cannot extract anything. Check "
                        "if the associated job was successfully completed."
                    )
                else:
                    raise ValueError(
                        f"Data with keys {result.data().keys()} in result not handled."
                    )

        if job.job_type == JobType.STATE_VECTOR:
            vector = result.get_statevector()
            state_vector = StateVector(
                vector.data,  # pyright: ignore[reportGeneralTypeIssues]
                job.circuit.nb_qubits,
            )
            return Result(job, state_vector, 0, 0)
        elif job.job_type == JobType.SAMPLE:
            assert job.measure is not None
            counts = result.get_counts()
            data = [
                Sample(
                    bin_str=item, count=counts[item], nb_qubits=job.circuit.nb_qubits
                )
                for item in counts
            ]
            return Result(job, data, None, job.measure.shots)
        else:
            raise NotImplementedError(f"{job.job_type} not handled.")

    else:
        raise NotImplementedError(f"Result type {type(result)} not handled")


@typechecked
def get_result_from_ibm_job_id(job_id: str) -> Result:
    """
    Retrieve from IBM remote platform and parse the result of the job_id given in parameter.
    If the job is still running, we wait (blocking) until it is DONE.

    Args:
        job_id: Id of the remote IBM job.

    Returns:
        A Result containing the formatted result of the executed ibm job.
    """
    # search for job id in the connector given in parameter first
    # if not found, try with IBMProvider, then QiskitRuntimeService
    # if not found, raise an error
    connector = get_IBMProvider()
    ibm_job = (
        connector.retrieve_job(job_id)
        if job_id in [job.job_id() for job in connector.jobs()]
        else None
    )
    if ibm_job is None:
        connector = get_QiskitRuntimeService()
        ibm_job = (
            connector.job(job_id)
            if job_id in [job.job_id() for job in connector.jobs()]
            else None
        )
    if ibm_job is None:
        raise IBMRemoteExecutionError(
            f"Job with id {job_id} was not found on this account."
        )

    status = ibm_job.status()
    if status in [IBM_JobStatus.CANCELLED, IBM_JobStatus.ERROR]:
        raise IBMRemoteExecutionError(
            f"Trying to retrieve an IBM result for a job in status {status.name}"
        )

    # If the job is finished, it will get the result, if still running it is block until it finishes
    result = ibm_job.result()
    backend = ibm_job.backend()
    assert isinstance(backend, (BackendV1, BackendV2))
    ibm_device = IBMDevice(backend.name)

    return extract_result(result, None, ibm_device, ibm_job)



import pytest

import numpy as np
from sympy import Expr, symbols

from mpqp.gates import *
from mpqp.tools.maths import matrix_eq, cos, sin, exp, I, pi
from mpqp.tools.generics import Matrix

theta: Expr
k: Expr
theta, k = symbols("θ k")  # pyright: ignore[reportAssignmentType]
c, s, e = cos(theta), sin(theta), exp(1.0 * I * theta)
c2, s2, e2 = (
    cos(theta / 2),  # pyright: ignore[reportOperatorIssue]
    sin(theta / 2),  # pyright: ignore[reportOperatorIssue]
    exp(1.0 * I * theta / 2),  # pyright: ignore[reportOperatorIssue]
)


@pytest.mark.parametrize(
    "angle, result_matrix",
    [
        (0, np.eye(2)),
        (np.pi, np.diag([1, -1])),
        (np.pi / 3, np.diag([1, 0.5 + 0.8660254j])),
        (theta, np.diag([1, e])),  # pyright: ignore
    ],
)
def test_P(angle: float, result_matrix: Matrix):
    assert matrix_eq(P(angle, 0).to_matrix(), result_matrix)


@pytest.mark.parametrize(
    "theta, phi, gamma, result_matrix",
    [
        (0, 0, 0, np.eye(2)),
        (0, 0, np.pi, np.diag([1, -1])),
        (0, np.pi, 0, np.diag([1, -1])),
        (np.pi, 0, 0, np.array([[0, -1], [1, 0]])),
        (
            2.75,
            1.23,
            3.08,
            np.array(
                [
                    [0.19454771 + 0.0j, 0.97903306 - 0.06037761j],
                    [0.32785147 + 0.92448072j, -0.07618831 - 0.1790088j],
                ]
            ),
        ),
        (
            theta,
            0,
            0,
            np.array([[c2, -1.0 * s2], [1.0 * s2, 1.0 * c2]]),  # pyright: ignore
        ),
    ],
)
def test_U(theta: float, phi: float, gamma: float, result_matrix: Matrix):
    assert matrix_eq(U(theta, phi, gamma, 0).to_matrix(), result_matrix)


@pytest.mark.parametrize(
    "angle, result_matrix",
    [
        (0, np.eye(2)),
        (np.pi, np.array([[0, -1j], [-1j, 0]])),
        (
            np.pi / 5,
            np.array(
                [
                    [0.95105652 + 0.0j, 0.0 - 0.30901699j],
                    [0.0 - 0.30901699j, 0.95105652 + 0.0j],
                ]
            ),
        ),
        (
            theta,
            np.array([[c2, -1j * s2], [-1j * s2, c2]]),  # pyright: ignore
        ),
    ],
)
def test_Rx(angle: float, result_matrix: Matrix):
    assert matrix_eq(Rx(angle, 0).to_matrix(), result_matrix)


@pytest.mark.parametrize(
    "angle, result_matrix",
    [
        (0, np.eye(2)),
        (np.pi, np.array([[0, -1], [1, 0]])),
        (
            np.pi / 5,
            np.array([[[0.95105652, -0.30901699], [0.30901699, 0.95105652]]]),
        ),
        (
            theta,
            np.array([[c2, -s2], [s2, c2]]),
        ),
    ],
)
def test_Ry(angle: float, result_matrix: Matrix):
    assert matrix_eq(Ry(angle, 0).to_matrix(), result_matrix)


@pytest.mark.parametrize(
    "angle, result_matrix",
    [
        (0, np.eye(2)),
        (np.pi, np.diag([-1j] * 2)),
        (np.pi / 5, np.diag([0.95105652 - 0.30901699j] * 2)),
        (theta, np.diag([1 / e2] * 2)),  # pyright: ignore
    ],
)
def test_Rz(angle: float, result_matrix: Matrix):
    assert matrix_eq(Rz(angle, 0).to_matrix(), result_matrix)


@pytest.mark.parametrize(
    "angle_bin_pow, result_matrix",
    [
        (0, np.diag([1, 1])),
        (1, np.diag([1, -1])),
        (5, np.diag([1, np.exp(1j * np.pi / 2**4)])),
        (k, np.diag([1, exp(1.0 * 2 ** (1 - k) * I * pi)])),  # pyright: ignore
    ],
)
def test_Rk(angle_bin_pow: int, result_matrix: Matrix):
    assert matrix_eq(Rk(angle_bin_pow, 0).to_matrix(), result_matrix)


@pytest.mark.parametrize(
    "angle_bin_pow, result_matrix",
    [
        (0, np.diag([1, 1, 1, 1])),
        (1, np.diag([1, 1, 1, -1])),
        (5, np.diag([1, 1, 1, np.exp(1j * np.pi / 2**4)])),
        (k, np.diag([1, 1, 1, exp(1.0 * 2 ** (1 - k) * I * pi)])),  # pyright: ignore
    ],
)
def test_CRk(angle_bin_pow: int, result_matrix: Matrix):
    assert matrix_eq(CRk(angle_bin_pow, 0, 1).to_matrix(), result_matrix)



import numpy as np
import pytest

from mpqp import QCircuit
from mpqp.gates import H, Rx
from mpqp.measures import ExpectationMeasure, Observable
from mpqp.execution import adjust_measure
from mpqp.tools.maths import matrix_eq


@pytest.mark.parametrize(
    "measure_targets, circuit, nb_ids_before, nb_ids_after",
    [
        ([0, 1], QCircuit([H(0), Rx(1.76, 2)]), 0, 1),
        ([1, 2], QCircuit([H(0), Rx(1.76, 2)]), 1, 0),
        ([1, 2], QCircuit([H(0), Rx(1.76, 3)]), 1, 1),
    ],
)
def test_adjust_measure(
    measure_targets: list[int],
    circuit: QCircuit,
    nb_ids_before: int,
    nb_ids_after: int,
):
    obs_matrix = np.array(
        [
            [0.63, 0.5, 1, 1],
            [0.5, 0.82, 1, 1],
            [1, 1, 1, 0.33],
            [1, 1, 0.33, 0.3],
        ],
    )
    measure = ExpectationMeasure(measure_targets, Observable(obs_matrix))
    adjusted_observable_matrix = np.kron(
        np.kron(
            np.eye(2**nb_ids_before, dtype=np.complex64), measure.observable.matrix
        ),
        np.eye(2**nb_ids_after),
    )
    assert matrix_eq(
        adjust_measure(measure, circuit).observable.matrix, adjusted_observable_matrix
    )



from __future__ import annotations
from typing import Literal
import dotenv

# Configuration file for the Sphinx documentation builder.
from sphinx.application import Sphinx
from sphinx.highlighting import PygmentsBridge
from pygments.formatters.latex import LatexFormatter

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys

sys.path.insert(0, os.path.abspath("../"))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "MPQP"
copyright = "2024, ColibrITD"
package_name = ""
author = "ColibrITD"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx_rtd_dark_mode",
    "sphinx_copybutton",
]
default_dark_mode = True
autodoc_typehints = "description"
autodoc_type_aliases = {"Matrix": "Matrix", "AvailableDevice": "AvailableDevice"}
simplify_optional_unions = True
typehints_defaults = "comma"
dotenv.load_dotenv()
sphinx_github_changelog_token = os.getenv("SPHINX_GITHUB_CHANGELOG_TOKEN")
if sphinx_github_changelog_token is not None:
    extensions.append("sphinx_github_changelog")

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# The reST default role (used for this markup: `text`) to use for all
# documents.
default_role = "math"

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = False

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False

pythonversion = sys.version.split(" ")[0]
# Python and Sage trac ticket shortcuts. For example, :trac:`7549` .
extlinks = {
    "python": ("https://docs.python.org/release/" + pythonversion + "/%s", ""),
    "trac": ("http://trac.sagemath.org/%s", "trac ticket #"),
    "wikipedia": ("https://en.wikipedia.org/wiki/%s", "Wikipedia article "),
    "arxiv": ("http://arxiv.org/abs/%s", "Arxiv "),
    "oeis": ("https://oeis.org/%s", "OEIS sequence "),
    "doi": ("https://dx.doi.org/%s", "doi:"),
    "mathscinet": ("http://www.ams.org/mathscinet-getitem?mr=%s", "MathSciNet "),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
# html_theme = "alabaster"
# html_theme = "furo"  # pip install furo
# html_theme = "sphinx_rtd_theme"  # pip install sphinx-rtd-theme

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# These paths are either relative to html_static_path
# or fully qualified paths (eg. https://...)
html_css_files = ["custom.css"]
# html_css_files = []

html_js_files = ["custom.js"]

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".

# A shorter title for the navigation bar.  Default is the same as html_title.
html_short_title = "MPQP"

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "resources/mpqp_test_logo.png"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "resources/mpqp_favicon.png"

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
# html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = package_name + "doc"

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    "preamble": "",
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        "index",
        package_name + ".tex",
        "Documentation of " + package_name,
        author,
        "manual",
    ),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [("index", package_name, package_name + " documentation", [author], 1)]

# If true, show URL addresses after external links.
# man_show_urls = False


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        "index",
        package_name,
        package_name + " documentation",
        author,
        package_name,
        project,
        "Miscellaneous",
    ),
]


# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
# texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
# texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
# texinfo_no_detailmenu = False


# This is to make the verbatim font smaller;
# Verbatim environment is not breaking long lines


class CustomLatexFormatter(LatexFormatter):  # type: ignore
    """Adds custom options to LaTeX outputs"""

    def __init__(self, **options):  # type: ignore
        super(CustomLatexFormatter, self).__init__(**options)
        self.verboptions = r"formatcom=\footnotesize"


PygmentsBridge.latex_formatter = CustomLatexFormatter

latex_elements[
    "preamble"
] += r"""
% One-column index
\makeatletter
\renewenvironment{theindex}{
  \chapter*{\indexname}
  \markboth{\MakeUppercase\indexname}{\MakeUppercase\indexname}
  \setlength{\parskip}{0.1em}
  \relax
  \let\item\@idxitem
}{}
\makeatother
\renewcommand{\ttdefault}{txtt}
"""

autodoc_default_options = {
    "members": None,
    "undoc-members": None,
    "show-inheritance": None,
    # 'private-members': None,
    # 'special-members': '__init__',
    # 'member-order': 'bysource',
    # 'exclude-members': '__weakref__'
}

autodoc_member_order = "groupwise"

WhatType = Literal["module", "class", "exception", "function", "method", "attribute"]
Options = Literal["inherited_members", "undoc_members", "show_inheritance", "no-index"]
OptionsType = dict[Options, str]


def maybe_skip_member(
    app: Sphinx,
    what: WhatType,
    name: str,
    obj: object,
    skip: bool,
    options: OptionsType,
):
    return skip or "M-TODO" in str(obj.__doc__)


def setup(app: Sphinx):
    app.connect("autodoc-skip-member", maybe_skip_member)



import os
from pathlib import Path
import re
from os.path import splitext
from enum import Enum

from warnings import warn
from textwrap import dedent
from anytree import Node, PreOrderIter
from typeguard import typechecked

from mpqp.tools.errors import InstructionParsingError


class Instr(Enum):
    """Special instruction of which the definition needs to included in the
    header of the file."""

    STD_LIB = 0
    CSX = 1
    U0 = 2
    CU3 = 3
    SXDG = 4
    RZZ = 5
    RXX = 6
    RCCX = 7
    RC3X = 8
    C3X = 9
    C4X = 10
    C3SQRTX = 11
    OQASM2_ALL_STDGATES = 12
    OQASM3_ALL_STDGATES = 13
    BRAKET_CUSTOM_INCLUDE = 14


std_gates_2 = [
    "cu3",
    "csx",
    "sxdg",
    "u0",
    "rxx",
    "rzz",
    "rccx",
    "rc3x",
    "c3x",
    "c3sqrtx",
    "c4x",
]
std_gates_2_3 = ["u", "swap", "cswap", "cp"]
std_gates_3 = [
    "u1",
    "u2",
    "u3",
    "cx",
    "CX",
    "id",
    "x",
    "y",
    "z",
    "h",
    "s",
    "sdg",
    "t",
    "tdg",
    "rx",
    "ry",
    "rz",
    "cz",
    "cy",
    "ch",
    "ccx",
    "crx",
    "cry",
    "crz",
    "cu",
    "p",
    "cphase",
    "phase",
    "sx",
]


@typechecked
def qasm_code(instr: Instr) -> str:
    """Return the string corresponding of the declaration of the instruction in
    parameter. It is also used to return the whole standard library string when
    we hard include it.

    Args:
        instr: Instr for which we want the corresponding OpenQASM code.

    Returns:
        OpenQASM definition of ``instr``.
    """

    # 3M-TODO: if run from outside the mpqp folder, the following line will fuck things up, fix it
    headers_folder = os.path.dirname(__file__) + "/header_codes/"
    special_file_names = {
        Instr.OQASM2_ALL_STDGATES: "qelib1.inc",
        Instr.OQASM3_ALL_STDGATES: "stdgates.inc",
        Instr.BRAKET_CUSTOM_INCLUDE: "braket_custom_include.inc",
    }

    if instr in special_file_names:
        file_name = special_file_names[instr]
    else:
        file_name = instr.name.lower() + ".qasm"

    with open(headers_folder + file_name, "r") as f:
        return f.read()


@typechecked
def parse_openqasm_2_file(code: str) -> list[str]:
    """Splits a complete OpenQASM2 program into individual instructions.

    Args:
        code: The complete OpenQASM 2.0 program, we do not check for correct syntax, it is assumed that the code is
            well formed.

    Returns:
        List of instructions.
    """
    # 6M-TODO: deal with comments, for the moment we remove them all

    # removing comment
    cleaned_code = "".join([loc.split("//")[0] for loc in code.split("\n")])

    cleaned_code = cleaned_code.replace("\t", " ").strip()

    # gate must be registered as a single instruction
    gate_matches = list(re.finditer(r"gate .*?}", cleaned_code))
    # we start with adding all the instructions up to the first gate definition
    sanitized_start = (
        cleaned_code[: gate_matches[0].span()[0]] if gate_matches else cleaned_code
    )
    instructions = sanitized_start.split(";")

    # we then add turn by turn the following gate definition, all the
    # instructions between it and the next one, and so on...
    for i in range(len(gate_matches)):
        instructions.append(cleaned_code[slice(*(gate_matches[i].span()))])
        instructions.extend(
            cleaned_code[
                gate_matches[i].span()[1] : (
                    None
                    if i == len(gate_matches) - 1
                    else gate_matches[i + 1].span()[0]
                )
            ].split(";")
        )

    return list(filter(lambda i: i != "", instructions))


@typechecked
def convert_instruction_2_to_3(
    instr: str,
    included_instr: set[Instr],
    included_tree_current_node: Node,
    defined_gates: set[str],
    path_to_main: str = ".",
) -> tuple[str, str]:
    """Some instructions changed name from QASM 2 to QASM 3, also the way to
    import files changed slightly. This function operates those changes on a
    single instruction.

    Args:
        instr: Instruction to be upgraded.
        included_instr: Some instructions need new imports, in order to keep
            track of which instruction are already.
        imported in the overall scope, a dictionary of already included
            instructions is passed and modified along.
        included_tree_current_node: Current Node in the file inclusion tree.
        defined_gates: Set of custom gates already defined.
        path_to_main: Path to the main folder from which include paths are
            described.

    Returns:
        The upgraded instruction and the potential code to add in the header as
        the second element.
    """

    def add_std_lib():
        """Add the instruction of including the standard library of OpenQASM3
        code if it is not already done"""
        if Instr.STD_LIB not in included_instr:
            included_instr.add(Instr.STD_LIB)
            to_add = qasm_code(Instr.STD_LIB)
        else:
            to_add = ""
        return to_add

    header_code = ""
    instructions_code = ""

    instr_name = instr.split(" ")[0].split("(")[0]

    # If the line is the OpenQASM header
    if instr.startswith("OPENQASM 2.0"):
        header_code += "OPENQASM 3.0;\n"
    elif instr_name == "include":
        path = instr.split(" ")[-1].strip("'\"")
        if path != "qelib1.inc":
            if is_path_in_ancestors(path, included_tree_current_node):
                raise RuntimeError("Circular dependency detected.")
            # Convert the file included, add it to the inclusion tree,
            # and create a new file and include it in the converted code
            if not is_path_in_tree(path, included_tree_current_node):
                with open(f"{path_to_main}/{path}", "r") as f:
                    child = Node(path, parent=included_tree_current_node)
                    converted_content = open_qasm_2_to_3(
                        f.read(), child, path_to_main, defined_gates
                    )
                new_path = splitext(path)[0] + "_converted" + splitext(path)[1]
                with open(f"{path_to_main}/{new_path}", "w") as f:
                    f.write(converted_content)
                header_code += f"include '{new_path}';\n"
    elif instr_name in {"qreg", "creg"}:
        # classical and quantum bits have the same structure
        # `qreg <name>[<size>];` -> `qubit[<size>] <name>;`
        m = re.match(r"(\w)reg\s+(.+?)\[(\d+)\]", instr)
        if m is None:
            raise InstructionParsingError("On instruction: " + instr)
        bit_type_prefix = "qu" if m.group(1) == "q" else ""
        instructions_code += f"{bit_type_prefix}bit[{m.group(3)}] {m.group(2)};\n"
    elif instr_name == "measure":
        # `measure <q_name[+reg]?> -> <c_name[+reg]?>;` -> `<c_name[+reg]?> = measure <q_name[+reg]?>;`
        m = re.match(r"measure\s+(.+?)\s+->\s+(.+)", instr)
        if m is None:
            raise InstructionParsingError("On instruction: " + instr)
        instructions_code += f"{m.group(2)} = measure {m.group(1)};\n"
    elif instr_name in {"reset", "barrier"}:
        instructions_code += instr + ";\n"
    elif instr_name.lower() == "u":
        warn(
            dedent(
                """OpenQASMTranslationWarning: 
                There is a phase e^(i(a+c)/2) difference between U(a,b,c) gate in 2.0 and 3.0.
                We handled that for you by adding the extra phase at the right place. 
                Be careful if you want to create a control gate from this circuit/gate, the
                phase can become non-global."""
            )
        )
        header_code += add_std_lib()
        instructions_code += "u3" + instr[1:] + ";\n"
    elif instr_name == "cu1":
        header_code += add_std_lib()
        instructions_code += "cp" + instr[3:] + ";\n"
    elif instr_name in std_gates_3 + std_gates_2_3 + std_gates_2:
        instructions_code += instr + ";\n"
        header_code += add_std_lib()
        new_instr = (
            [instruc for instruc in Instr if instr_name == instruc.name.lower()][0]
            if instr_name in std_gates_2
            else None
        )
        if new_instr is not None and new_instr not in included_instr:
            included_instr.add(new_instr)
            header_code += qasm_code(new_instr)
    elif instr_name == "gate":
        defined_gates.add(instr.split()[1])
        g_string = instr.split("{")[0] + "{\n"
        g_instructions = filter(
            lambda i: not re.fullmatch(r"\s*", i),
            instr.split("{")[1].split("}")[0].split(";"),
        )
        for instruction in g_instructions:
            instruction = instruction.strip()
            i_code, h_code = convert_instruction_2_to_3(
                instruction,
                included_instr,
                included_tree_current_node,
                defined_gates,
                path_to_main,
            )
            g_string += " " * 4 + i_code
            header_code += h_code
        instructions_code += g_string + "}\n"
    elif instr_name == "if":
        if_statement = instr.split(")")[0] + ")"
        nested_instr = ")".join(instr.split(")")[1:])
        i_code, h_code = convert_instruction_2_to_3(
            nested_instr,
            included_instr,
            included_tree_current_node,
            defined_gates,
            path_to_main,
        )
        instructions_code += if_statement + i_code + ";\n"
        header_code += h_code
    elif instr_name == "opaque":
        raise NotImplementedError("opaque exports not handled yet")
    else:
        gate = instr.split()[0]
        if gate == "ctrl":
            gate = instr.split()[2]
        if gate not in defined_gates:
            raise ValueError(f"Gates undefined at the time of usage: {gate}")
        if len(instr) != 0:
            instructions_code += instr + ";\n"

    return instructions_code, header_code


@typechecked
def open_qasm_2_to_3(
    code: str,
    included_tree_current_node: Node = Node("initial_code"),
    path_to_file: str = ".",
    defined_gates: set[str] = set(),
) -> str:
    """Converts an OpenQASM code from version 2.0 and 3.0.

    It is a partial conversion (mainly circuit structure) for helping building
    temporary bridges between different platforms that use different versions.

    Args:
        code: String containing the OpenQASM 2.0 code and instructions.
        included_tree_current_node: Current Node in the file inclusion tree.
        path_to_file: Path to the location of the file from which the code is
            coming (useful for locating imports).
        defined_gates: Set of custom gates already defined.

    Returns:
        Converted OpenQASM code in the 3.0 version.

    Example:
        >>> qasm2_str = '''\\
        ... OPENQASM 2.0;
        ... qreg q[2];
        ... creg c[2];
        ... h q[0];
        ... cx q[0],q[1];
        ... measure q[0] -> c[0];
        ... measure q[1] -> c[1];
        ... '''
        >>> open_qasm_2_to_3(qasm2_str)
        '''OPENQASM 3.0;
        include 'stdgates.inc';
        qubit[2] q;
        bit[2] c;
        h q[0];
        cx q[0],q[1];
        c[0] = measure q[0];
        c[1] = measure q[1];
        '''
    """

    header_code = ""
    instructions_code = ""

    instructions = parse_openqasm_2_file(code)

    included_instructions = set()

    defined_gates.update(std_gates_2_3 + std_gates_3)

    for instr in instructions:
        i_code, h_code = convert_instruction_2_to_3(
            instr,
            included_instructions,
            included_tree_current_node,
            defined_gates,
            path_to_file,
        )
        header_code += h_code
        instructions_code += i_code

    target_code = header_code + "\n" + instructions_code

    return target_code


@typechecked
def open_qasm_file_conversion_2_to_3(path: str) -> str:
    """Converts an OpenQASM code in a file from version 2.0 and 3.0.

    It is a partial conversion (mainly circuit structure) for helping building
    temporary bridges between different platforms that use different versions.

    Args:
        path: Path to the file containing the OpenQASM 2.0 code, and eventual
            imports.

    Returns:
        Converted OpenQASM code in the 3.0 version.

    Example:
        >>> example_dir = "example/qasm_files/"
        >>> with open(example_dir + "main.qasm", "r") as f:
        ...     print(f.read())
        '''OPENQASM 2.0;
        include "include1.qasm";
        include "include2.qasm";
        qreg q[2];
        creg c[2];
        h q[0];
        cx q[0],q[1];
        gate2 q[0];
        gate3 q[0], q[1];
        measure q[0] -> c[0];
        measure q[1] -> c[1];'''
        >>> print(open_qasm_file_conversion_2_to_3(example_dir + "main.qasm"))
        '''OPENQASM 3.0;
        include 'include1_converted.qasm';
        include 'include2_converted.qasm';
        include 'stdgates.inc';
        qubit[2] q;
        bit[2] c;
        h q[0];
        cx q[0],q[1];
        gate2 q[0];
        gate3 q[0], q[1];
        c[0] = measure q[0];
        c[1] = measure q[1];'''
        >>> with open(example_dir + "include1_converted.qasm", "r") as f:
        ...     print(f.read())
        '''OPENQASM 3.0;
        include 'stdgates.inc';
        gate gate2 a {
            u3(pi, -pi/2, pi/2) a;
        }'''
        >>> with open(example_dir + "include2_converted.qasm", "r") as f:
        ...     print(f.read())
        '''OPENQASM 3.0;
        include 'stdgates.inc';
        gate gate3 a, b {
            u3(0, -pi/2, pi/3) a;
            cz a, b;
        }'''
    """

    with open(path, "r") as f:
        code = f.read()
        return open_qasm_2_to_3(code, Node(path), str(Path(path).parent))


@typechecked
def open_qasm_hard_includes(
    code: str,
    included_files: set[str],
    path_to_file: str = "./",
    is_openqasm_header_included: bool = False,
) -> str:
    r"""Converts an OpenQASM code (2.0 and 3.0) to use no includes, but writes
    every instruction in previously included files, directly in the code
    returned.

    Example:
        >>> examples_folder = "tests/qasm/qasm_examples"
        >>> filename = examples_folder + "/with_include.qasm"
        >>> with open(filename) as f:
        ...     print(open_qasm_hard_includes(f.read(), {filename}).strip("\n"))
        '''gate csx a, b {
            ctrl @ sx a, b;
        }'''

    Args:
        code: String containing the OpenQASM code and instructions.
        included_files: The set of files already included, used to avoid
            duplicate imports and circular dependencies. This set should be
            initialized with the name of the root file you started with.
        path_to_file: Path used to localize files that are included.
        is_openqasm_header_included: Boolean used to only include once the
            OpenQASM header.

    Returns:
        Include-less OpenQASM code.
    """
    lines = code.split("\n")
    converted_code = []

    for line in lines:
        if "include" in line:
            line_array = line.split()

            file_name = line_array[line_array.index("include") + 1].strip(";'\"")
            if file_name not in included_files:
                included_files.add(file_name)
                if file_name in {"qelib1.inc"}:
                    converted_code.append(qasm_code(Instr.OQASM2_ALL_STDGATES))
                elif file_name in {"stdgates.inc"}:
                    converted_code.append(qasm_code(Instr.OQASM3_ALL_STDGATES))
                elif file_name in {"braket_custom_include.inc"}:
                    converted_code.append(qasm_code(Instr.BRAKET_CUSTOM_INCLUDE))
                else:
                    with open(path_to_file + file_name, "r") as f:
                        converted_code.append(
                            open_qasm_hard_includes(
                                f.read(),
                                included_files,
                                path_to_file,
                                is_openqasm_header_included,
                            )
                        )
        elif line.startswith("OPENQASM "):
            if not is_openqasm_header_included:
                converted_code.append(line)
                is_openqasm_header_included = True
        else:
            if not line.startswith("//"):
                converted_code.append(line)

    return "\n".join(converted_code)


@typechecked
def is_path_in_tree(path: str, any_node: Node):
    """Checks if a path is already present in a node of the tree represented by
    one of his nodes (in parameter).

    Args:
        path: Element to search in the tree.
        any_node: One node of the tree on which we want to search.

    Returns:
        ``True`` if the path is the name of one of the nodes of the tree.
    """
    return any([path in node.name for node in PreOrderIter(any_node.root)])


@typechecked
def is_path_in_ancestors(path: str, node: Node):
    """Checks if a path is already present in an ancestor of the node in
    parameter.

    Args:
        path: Element to search in the tree.
        node: The node for which we want to search in his ancestors.

    Returns:
        ``True`` if the path is the name of one of the ancestors of the node/
    """
    return any([path in node.name for node in node.ancestors])



"""For now, our minimizer is a wrapper around ``scipy``'s minimizer. The
:class:`Optimizer` enum lists all the methods validated with the rest of the
library."""

from enum import Enum, auto


class Optimizer(Enum):
    """Enum used to select the optimizer for the VQA."""

    BFGS = auto()
    COBYLA = auto()
    CMAES = auto()



from typing import Optional

import numpy as np
from typeguard import typechecked
from statistics import mean

from mpqp.core.instruction.measurement import ComputationalBasis
from mpqp.core.instruction.measurement.basis_measure import BasisMeasure
from mpqp.core.instruction.measurement.expectation_value import (
    ExpectationMeasure,
    Observable,
)
from mpqp.execution.devices import ATOSDevice
from ..connection.qlm_connection import get_QLMaaSConnection
from ..job import Job, JobType, JobStatus
from ..result import Result, Sample, StateVector
from mpqp.qasm import qasm2_to_myqlm_Circuit
from mpqp import QCircuit, Language
from ...tools.errors import QLMRemoteExecutionError

from qat.qlmaas.result import AsyncResult
from qat.core.contexts import QPUContext
from qat.core.qpu.qpu import QPUHandler
from qat.pylinalg import PyLinalg
from qat.clinalg.qpu import CLinalg
from qat.core.wrappers.observable import Observable as QLM_Observable
from qat.plugins.observable_splitter import ObservableSplitter
from qat.core.wrappers.result import Result as QLM_Result
from qat.core.wrappers.circuit import Circuit
from qat.core.wrappers.job import Job as JobQLM
from qat.comm.qlmaas.ttypes import JobStatus as QLM_JobStatus, QLMServiceException


@typechecked
def get_local_qpu(device: ATOSDevice) -> QPUHandler:
    """
    Returns the myQLM local QPU associated with the ATOSDevice given in parameter.

    Args:
        device: ATOSDevice referring to the myQLM local QPU.

    Raises:
        ValueError
    """
    if device.is_remote():
        raise ValueError("Excepted a local device, not the remote QLM")
    if device == ATOSDevice.MYQLM_PYLINALG:
        return PyLinalg()
    return CLinalg()


@typechecked
def generate_state_vector_job(
    myqlm_circuit: Circuit, device: ATOSDevice = ATOSDevice.MYQLM_PYLINALG
) -> tuple[JobQLM, QPUContext]:
    """Generates a myQLM job from the myQLM circuit and selects the right myQLM
    QPU to run on it.

    Args:
        myqlm_circuit: MyQLM circuit of the job.
        device: ATOSDevice on which the user wants to run the job (to know if
            the run in local of remote).

    Returns:
        A myQLM Job and the right myQLM QPUHandler on which it will be submitted.
    """

    if device.is_remote():
        get_QLMaaSConnection()
        from qlmaas.qpus import LinAlg  # type: ignore

        qpu = LinAlg()
    else:
        qpu = get_local_qpu(device)

    return myqlm_circuit.to_job(job_type="SAMPLE"), qpu


@typechecked
def generate_sample_job(myqlm_circuit: Circuit, job: Job) -> tuple[JobQLM, QPUContext]:
    """Generates a myQLM job from the myQLM circuit and job sample info (target,
    shots, ...), and selects the right myQLM QPU to run on it.

    Args:
        myqlm_circuit: MyQLM circuit of the job.
        job: Original mpqp job used to generate the myQLM job.

    Returns:
        A myQLM Job and the right myQLM QPUHandler on which it will be submitted.
    """

    assert job.measure is not None

    myqlm_job = myqlm_circuit.to_job(
        job_type="SAMPLE",
        qubits=job.measure.targets,
        nbshots=job.measure.shots,
    )

    if job.device.is_remote():
        get_QLMaaSConnection()
        from qlmaas.qpus import LinAlg  # type: ignore

        qpu = LinAlg()
    else:
        assert isinstance(job.device, ATOSDevice)
        qpu = get_local_qpu(job.device)

    return myqlm_job, qpu


@typechecked
def generate_observable_job(
    myqlm_circuit: Circuit, job: Job
) -> tuple[JobQLM, QPUContext]:
    """Generates a myQLM job from the myQLM circuit and observable, and selects
    the right myQLM QPU to run on it.

    Args:
        myqlm_circuit: MyQLM circuit of the job.
        job: Original mpqp job used to generate the myQLM job.

    Returns:
        A myQLM Job and the right myQLM QPUHandler on which it will be submitted.
    """

    assert job.measure is not None and isinstance(job.measure, ExpectationMeasure)

    myqlm_job = myqlm_circuit.to_job(
        job_type="OBS",
        observable=QLM_Observable(
            job.measure.nb_qubits, matrix=job.measure.observable.matrix
        ),
        nbshots=job.measure.shots,
    )
    if job.device.is_remote():
        get_QLMaaSConnection()
        from qlmaas.qpus import LinAlg  # type: ignore

        qpu = LinAlg()
    else:
        assert isinstance(job.device, ATOSDevice)
        qpu = ObservableSplitter() | get_local_qpu(job.device)

    return myqlm_job, qpu


@typechecked
def extract_state_vector_result(
    myqlm_result: QLM_Result,
    job: Optional[Job] = None,
    device: ATOSDevice = ATOSDevice.MYQLM_PYLINALG,
) -> Result:
    """Constructs a Result from the result given by the myQLM/QLM run in state
    vector mode.

    Args:
        myqlm_result: Result returned by myQLM/QLM after running of the job.
        job: Original mpqp job used to generate the run. Used to retrieve more
            easily info to instantiate the result.
        device: ATOSDevice on which the job was submitted. Used to know if the
            run was remote or local.

    Returns:
        A Result containing the result info extracted from the myQLM/QLM
        statevector result.
    """
    if job is None:
        nb_qubits = (
            myqlm_result.qregs[0].length
            if device.is_remote()
            else sum(len(qreg.qbits) for qreg in myqlm_result.data.qregs)
        )
        job = Job(JobType.STATE_VECTOR, QCircuit(nb_qubits), device, None)
    else:
        nb_qubits = job.circuit.nb_qubits

    nb_states = 2**nb_qubits
    amplitudes = np.zeros(nb_states, np.complex64)
    probas = np.zeros(nb_states, np.float32)
    for sample in myqlm_result:
        amplitudes[sample._state] = sample.amplitude
        probas[sample._state] = sample.probability

    return Result(job, StateVector(amplitudes, nb_qubits, probas), 0, 0)


@typechecked
def extract_sample_result(
    myqlm_result: QLM_Result,
    job: Optional[Job] = None,
    device: ATOSDevice = ATOSDevice.MYQLM_PYLINALG,
) -> Result:
    """Constructs a Result from the result given by the myQLM/QLM run in sample
    mode.

    Args:
        myqlm_result: Result returned by myQLM/QLM after running of the job.
        job: Original mpqp job used to generate the run. Used to retrieve more
            easily info to instantiate the result.
        device: ATOSDevice on which the job was submitted. Used to know if the
            run was remote or local.

    Returns:
        A Result containing the result info extracted from the myQLM/QLM sample
        result.
    """
    if job is None:
        assert isinstance(myqlm_result.qregs[0].length, int)
        nb_qubits = (
            myqlm_result.qregs[0].length
            if device.is_remote()
            else sum(len(qreg.qbits) for qreg in myqlm_result.raw_data.qregs)
        )
        nb_shots = int(myqlm_result.meta_data["nbshots"])
        job = Job(
            JobType.SAMPLE,
            QCircuit(nb_qubits),
            device,
            BasisMeasure(targets=list(range(nb_qubits)), shots=nb_shots),
        )
    else:
        nb_qubits = job.circuit.nb_qubits
        if job.measure is None:
            raise NotImplementedError("We cannot handle job without measure for now")
        nb_shots = job.measure.shots

    # we here take the average of errors over all samples
    error = mean([sample.err for sample in myqlm_result])
    samples = [
        Sample(
            nb_qubits,
            index=sample._state,
            probability=sample.probability,
            bin_str=str(sample.state)[1:-1],
        )
        for sample in myqlm_result
    ]

    return Result(job, samples, error, nb_shots)


@typechecked
def extract_observable_result(
    myqlm_result: QLM_Result,
    job: Optional[Job] = None,
    device: ATOSDevice = ATOSDevice.MYQLM_PYLINALG,
) -> Result:
    """Constructs a Result from the result given by the myQLM/QLM run in
    observable mode.

    Args:
        myqlm_result: Result returned by myQLM/QLM after running of the job.
        job: Original mpqp job used to generate the run. Used to retrieve more
            easily info to instantiate the result.
        device: ATOSDevice on which the job was submitted. Used to know if the
            run was remote or local.

    Returns:
        A Result containing the result info extracted from the myQLM/QLM
        observable result.
    """
    if job is None:
        if device.is_remote():
            nb_qubits = myqlm_result.data.qregs[0].length
        else:
            nb_qubits = sum(len(qreg.qbits) for qreg in myqlm_result.data.qregs)
        nb_shots = int(myqlm_result.meta_data["nbshots"])
        job = Job(
            JobType.OBSERVABLE,
            QCircuit(nb_qubits),
            device,
            ExpectationMeasure(
                targets=list(range(nb_qubits)),
                observable=Observable(
                    np.zeros((2**nb_qubits, 2**nb_qubits), dtype=np.complex64)
                ),
                shots=nb_shots,
            ),
        )
    else:
        if job.measure is None:
            raise NotImplementedError("We cannot handle job without measure for now")
        nb_shots = job.measure.shots

    return Result(job, myqlm_result.value, myqlm_result.error, nb_shots)


@typechecked
def extract_result(
    myqlm_result: QLM_Result,
    job: Optional[Job] = None,
    device: ATOSDevice = ATOSDevice.MYQLM_PYLINALG,
) -> Result:
    """
    Constructs a Result from the result given by the myQLM/QLM run.

    Args:
        myqlm_result: Result returned by myQLM/QLM after running of the job.
        job: Original mpqp job used to generate the run. Used to retrieve more
            easily info to instantiate the result.
        device: ATOSDevice on which the job was submitted. Used to know if the
            run was remote or local.

    Returns:
        A Result containing the result info extracted from the myQLM/QLM result.
    """

    if (job is None) or job.device.is_remote():
        if myqlm_result.value is None:
            if list(myqlm_result)[0].amplitude is None:
                job_type = JobType.SAMPLE
            else:
                job_type = JobType.STATE_VECTOR
        else:
            job_type = JobType.OBSERVABLE
    else:
        job_type = job.job_type

    if job_type == JobType.STATE_VECTOR:
        return extract_state_vector_result(myqlm_result, job, device)
    elif job_type == JobType.SAMPLE:
        return extract_sample_result(myqlm_result, job, device)
    else:
        return extract_observable_result(myqlm_result, job, device)


@typechecked
def job_pre_processing(job: Job) -> Circuit:
    """Extracts the myQLM circuit and check if ``job.type`` and ``job.measure``
    are coherent.

    Args:
        job: Mpqp job used to instantiate the myQLM circuit.

    Returns:
          The myQLM Circuit translated from the circuit of the job in parameter.
    """

    if (
        job.job_type == JobType.STATE_VECTOR
        and job.measure is not None
        and not isinstance(job.measure, BasisMeasure)
    ):
        raise ValueError("`STATE_VECTOR` jobs require basis measure to be run")
    if job.job_type == JobType.OBSERVABLE and not isinstance(
        job.measure, ExpectationMeasure
    ):
        raise ValueError("`OBSERVABLE` jobs require `ExpectationMeasure` to be run")

    myqlm_circuit = job.circuit.to_other_language(Language.MY_QLM)

    return myqlm_circuit


@typechecked
def run_atos(job: Job) -> Result:
    """Executes the job on the right ATOS device precised in the job in parameter.
    This function is not meant to be used directly, please use
    ``runner.run(...)`` instead.

    Args:
        job: Job to be executed.

    Returns:
        A Result after submission and execution of the job.
    """
    return run_myQLM(job) if not job.device.is_remote() else run_QLM(job)


@typechecked
def run_myQLM(job: Job) -> Result:
    """Executes the job on the local myQLM simulator. This function is not meant
    to be used directly, please use ``runner.run(...)`` instead.

    Args:
        job: Job to be executed.

    Returns:
        A Result after submission and execution of the job.
    """

    result = None
    myqlm_job = None
    myqlm_result = None
    qpu = None

    myqlm_circuit = job_pre_processing(job)

    if job.job_type == JobType.STATE_VECTOR:
        myqlm_job, qpu = generate_state_vector_job(myqlm_circuit)

    elif job.job_type == JobType.SAMPLE:
        assert isinstance(job.measure, BasisMeasure)
        if isinstance(job.measure.basis, ComputationalBasis):
            myqlm_job, qpu = generate_sample_job(myqlm_circuit, job)
        else:
            raise NotImplementedError(
                "Does not handle other basis than the ComputationalBasis for the moment"
            )

    elif job.job_type == JobType.OBSERVABLE:
        assert isinstance(job.measure, ExpectationMeasure)
        myqlm_job, qpu = generate_observable_job(myqlm_circuit, job)

    else:
        raise ValueError(f"Job type {job.job_type} not handled")

    job.status = JobStatus.RUNNING
    myqlm_result = qpu.submit(myqlm_job)

    # retrieving the results
    result = extract_result(myqlm_result, job, job.device)  # type: ignore

    job.status = JobStatus.DONE
    return result


@typechecked
def submit_QLM(job: Job) -> tuple[str, AsyncResult]:
    """Submits the job on the remote QLM machine. This function is not meant to
    be used directly, please use ``runner.submit(...)`` instead.

    Args:
        job: Job to be executed.

    Returns:
        The job_id and the AsyncResult of the submitted job.
    """

    myqlm_job = None
    qpu = None

    myqlm_circuit = job_pre_processing(job)

    if job.device == ATOSDevice.QLM_LINALG:
        if job.job_type == JobType.STATE_VECTOR:
            assert isinstance(job.device, ATOSDevice)
            myqlm_job, qpu = generate_state_vector_job(myqlm_circuit, job.device)

        elif job.job_type == JobType.SAMPLE:
            assert isinstance(job.measure, BasisMeasure)
            if isinstance(job.measure.basis, ComputationalBasis):
                myqlm_job, qpu = generate_sample_job(myqlm_circuit, job)
            else:
                raise NotImplementedError(
                    "Does not handle other basis than the ComputationalBasis for the moment"
                )

        elif job.job_type == JobType.OBSERVABLE:
            assert isinstance(job.measure, ExpectationMeasure)
            myqlm_job, qpu = generate_observable_job(myqlm_circuit, job)

        else:
            raise ValueError(f"Job type {job.job_type} not handled")

        job.status = JobStatus.RUNNING
        async_result = qpu.submit(myqlm_job)
        job_id = async_result.get_info().id
        job.id = job_id
    else:
        raise NotImplementedError(f"Device {job.device} not handled")

    return job_id, async_result


@typechecked
def run_QLM(job: Job) -> Result:
    """Submits the job on the remote QLM machine and waits for it to be done. This
    function is not meant to be used directly, please use ``runner.run(...)``
    instead.

    Args:
        job: Job to be executed.

    Returns:
        A Result after submission and execution of the job.
    """

    if not isinstance(job.device, ATOSDevice) or not job.device.is_remote():
        raise ValueError(
            "The job given is not a QLM one, so it cannot be handled by this "
            "function."
        )

    _, async_result = submit_QLM(job)
    qlm_result = async_result.join()

    return extract_result(qlm_result, job, job.device)


@typechecked
def get_result_from_qlm_job_id(job_id: str) -> Result:
    """Retrieve the result, described by the job_id in parameter, from the
    remote QLM and convert it into an mpqp result. If the job is still running,
    we wait (blocking) until it is DONE.

    Args:
        job_id: Id of the remote QLM job.

    Raises:
        QLMRemoteExecutionError

    """
    connection = get_QLMaaSConnection()

    try:
        qlm_job = connection.get_job(job_id)
    except QLMServiceException as e:
        raise QLMRemoteExecutionError(
            f"Job with id {job_id} not found.\nTrace: " + str(e)
        )

    status = qlm_job.get_status()
    if status in [
        QLM_JobStatus.CANCELLED,
        QLM_JobStatus.UNKNOWN_JOB,
        QLM_JobStatus.DELETED,
        QLM_JobStatus.FAILED,
        QLM_JobStatus.STOPPED,
    ]:
        raise QLMRemoteExecutionError(
            f"Trying to retrieve a QLM result for a job in status {status.name}"
        )
    elif status in [QLM_JobStatus.WAITING, QLM_JobStatus.RUNNING]:
        qlm_job.join()

    qlm_result: QLM_Result = qlm_job.get_result()

    return extract_result(qlm_result, None, ATOSDevice.QLM_LINALG)



from __future__ import annotations
from typing import Optional

from mpqp.core.instruction.gates import (
    KrausRepresentation,
    PauliDecomposition,
)
from mpqp.core.languages import Language

import numpy as np
import numpy.typing as npt

# MYQLM
# from qat.lang.AQASM import Program, H, PH, CNOT, SWAP, RX
from qat.quops.quantum_channels import QuantumChannelKraus

# QISKIT

# Import from Qiskit Aer noise module
from qiskit_aer.noise import depolarizing_error


class GateNoise:
    """
    A class used to apply Noise on a specific Gate

    Noise can be defined in several ways, and this class allows us apply a
    specific qubit quantum error to a selected gate

    Args:

    Attributes:


    Inspiration from MyQLM to be removed later
    # for each gate, we specify the noise
    # note that the values in this dictionary are lambda functions
    # with as many arguments as the gate's number of arguments
    gates_noise = {"H": lambda: noise,
               "CNOT": lambda: noise2,
               "RX": lambda _: noise}

    #This will go in the MPQPNoiseModel class
    hw_model = HardwareModel(gates_spec, gates_noise, idle_noise=None)

    """

    def __init__(
        self,
    ):
        # 3M-TODO : implement and comment
        pass


class GateNoiseCombination:
    """
    A class that allows to combine different types of Noise on a selected Gate(s)
    by using composition, tensor product, and tensor expansion and thus produces a
    new quantum error.


    Args:?
        matrix : unitary matrix representing the gate
        gate_combination : combination of gates (sum, product, ...) defining the gate
        kraus_operators : generalized Kraus representation of the gate
        pauli_decomposition : when it is possible, decomposition of the gate in the Pauli basis (I, X, Y, Z)
        nb_qubits : number of qubits of the gate defined

        more has to be added here, once it is thought through

    Attributes:
        _current_type (str): string describing which definition is currently used to define the gate


    """

    def __init__(self):
        # 3M-TODO : implement and comment
        self.pp = 1


class NoiseModules:
    """
    A class that contains function and methods for noise modeling.


    Args:
        matrix : unitary matrix representing the gate
        gate_combination : combination of gates (sum, product, ...) defining the gate
        kraus_operators : generalized Kraus representation of the gate
        pauli_decomposition : when it is possible, decomposition of the gate in the Pauli basis (I, X, Y, Z)
        nb_qubits : number of qubits of the gate defined

        more has to be added here, once it is thought through

    Attributes:
        _current_type (str): string describing which definition is currently used to define the gate


    """

    def __init__(self, nb_qubits: int, param: float, language: str):
        self._nb_qubits = nb_qubits
        self._language = language
        # put more

    def depolarizing_error(
        self, param: float, num_qubits: int, language: Language = Language.QISKIT
    ):
        """
        Args:
            param: depolarizing error parameter.
            num_qubits: the number of qubits for the error channel.
            language: ...

        Instantiating a noise model representing a depolarizing noise with probability 5%

        Example:
            For Qiskit:
            >>> error = depolarizing_error(0.05, 1)

        Example:
            For MyQLM:
            >>> prob = 0.05
            >>> X = np.array([[0, 1], [1, 0]])
            >>> Y = np.array([[0, -1j], [1j, 0]])
            >>> Z = np.array([[1, 0], [0, -1]])
            >>> kraus_ops = [np.sqrt(1-prob)*np.identity(2),
            >>> np.sqrt(prob/3)*X, np.sqrt(prob/3)*Y, np.sqrt(prob/3)*Z]
            >>> error = QuantumChannelKraus(kraus_ops)

        #How to connect?

        """

        # 3M-TODO : finish and comment

        if language == Language.QISKIT:
            error = depolarizing_error(param, num_qubits)
        elif language == Language.MY_QLM:
            X = np.array([[0, 1], [1, 0]])
            Y = np.array([[0, -1j], [1j, 0]], dtype=complex)
            Z = np.array([[1, 0], [0, -1]])
            kraus_ops = [
                np.sqrt(1 - param) * np.identity(2),
                np.sqrt(param / 3) * X,
                np.sqrt(param / 3) * Y,
                np.sqrt(param / 3) * Z,
            ]
            error = QuantumChannelKraus(kraus_ops)
        else:
            raise NotImplementedError(f"{language} not supported")

        return error

    def compose_noise(
        self, language: Language = Language.QISKIT
    ):  # can be also in MYQLM
        """
        Composes two noise gates

        Args:
            error1: ?
            error2:?
            noise_combi_type: bool
            etc?


        Returns: new_error
            ?

        """

        # here use mpqp tensor product function?
        # if noise_combi_type == noise_composition:
        #     error = error1.compose(error2)
        # elif noise_combi_type == noise_tensor_product:
        #     error = error1.tensor(error2)
        # elif noise_combi_type == noise_expand_product:
        #     error = error1.expand(error2)

        pass

    def krauss_error(
        self, language: Language = Language.QISKIT
    ):  # can be also in MYQLM
        """
        a general n-qubit CPTP error channel given as a list of Kraus matrices

        Args: mpqp_noise_operators (list[matrix]): Kraus matrices.
            ?

        Returns:
            The quantum error object - a Kraus quantum error channel.

        """
        pass



"""When we measure a quantum state, we project it in an orthonormal basis of the
associated Hilbert space. By default,
:class:`BasisMeasure<mpqp.core.instruction.measurement.basis_measure.BasisMeasure>`
operates in the computational basis, but you may want to measure the state in a
custom basis, like it can be the case in the Bell game. For this purpose, you
can use the :class:`Basis` class.

On the other hand, some common basis are available for you to use:
:class:`ComputationalBasis` and :class:`HadamardBasis`."""

from __future__ import annotations
from abc import abstractmethod
from functools import reduce
from typing import Optional

import numpy as np
import numpy.typing as npt
from typeguard import typechecked

from mpqp.tools.maths import matrix_eq, atol


@typechecked
class Basis:
    """Represents a basis of the Hilbert space used for measuring a qubit.

    Args:
        basis_vectors: List of vector composing the basis.
        nb_qubits: Number of qubits associated with this basis. If not
            specified, it will be automatically inferred from
            ``basis_vectors``'s dimensions.

    Example:
        >>> Basis([np.array([1,0]), np.array([0,-1])]).pretty_print()
        Basis: [
            [1 0],
            [ 0 -1]
        ]
    """

    def __init__(
        self,
        basis_vectors: list[npt.NDArray[np.complex64]],
        nb_qubits: Optional[int] = None,
    ):
        # 3M-TODO : add the possibility to give the symbols for the '0' and '1' of the custom basis. This should then
        #  appear in the Sample binary_representation of the basis state. For instance in the Hadamard basis, the
        #  symbols will be '+' and '-'. If the user wants '↑' and '↓' for his custom basis, when we print samples we
        #  would have something like:
        #  State: ↑↑↓, Index: 1, Count: 512, Probability: 0.512
        if len(basis_vectors) == 0:
            self.nb_qubits = nb_qubits
            self.basis_vectors = basis_vectors
            return
        if nb_qubits is None:
            nb_qubits = int(np.log2(len(basis_vectors[0])))

        if len(basis_vectors) != 2**nb_qubits:
            raise ValueError(
                "Incorrect number of vector for the basis: given "
                f"{len(basis_vectors)} but there should be {2**nb_qubits}"
            )
        if any(len(vector) != 2**nb_qubits for vector in basis_vectors):
            raise ValueError("All vectors of the given basis are not the same size")
        if any(
            abs(np.linalg.norm(vector) - 1)  # pyright: ignore[reportGeneralTypeIssues]
            > atol
            for vector in basis_vectors
        ):
            raise ValueError("All vectors of the given basis are not normalized")
        m = np.array([vector for vector in basis_vectors])
        if not matrix_eq(
            m.transpose().dot(m),
            np.eye(len(basis_vectors)),  # pyright: ignore[reportArgumentType]
        ):
            raise ValueError("The given basis is not orthogonal")

        self.nb_qubits = nb_qubits
        """See parameter description."""
        self.basis_vectors = basis_vectors
        """See parameter description."""

    def pretty_print(self):
        """Nicer print for the basis, with human readable formatting.

        Example:
            >>> Basis([np.array([1,0]), np.array([0,-1])]).pretty_print()
            Basis: [
                [ 1  0],
                [ 0 -1]
            ]
        """
        joint_vectors = ",\n    ".join(map(str, np.round(self.basis_vectors, 2)))
        print(f"Basis: [\n    {joint_vectors}\n]")

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.basis_vectors}, {self.nb_qubits})"


@typechecked
class VariableSizeBasis(Basis):
    """3M-TODO"""

    @abstractmethod
    def __init__(self, nb_qubits: Optional[int] = None):
        pass

    @abstractmethod
    def set_size(self, nb_qubits: int):
        """
        To allow the user to use a basis without having to specify the size
        (because implicitly the size should be the number of qubits of the
        circuit), we use this method, that will only be called once the
        circuit's size is definitive (i.e. at the last moment before the circuit
        is ran)

        Args:
            nb_qubits: number of qubits in the basis
        """
        pass

    def __repr__(self) -> str:
        return f"{type(self).__name__}()"


class ComputationalBasis(VariableSizeBasis):
    """Basis representing the computational basis, also called Z-basis or
    canonical basis.

    Args:
        nb_qubits: number of qubits of the space, if not given as input (for
            example if unknown at the moment of creation) ``set_size`` will have
            to be executed before the basis is used (in a measure for example).

    Examples:
        >>> ComputationalBasis(3).pretty_print()
        Basis: [
            [1.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j],
            [0.+0.j 1.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j],
            [0.+0.j 0.+0.j 1.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j],
            [0.+0.j 0.+0.j 0.+0.j 1.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j],
            [0.+0.j 0.+0.j 0.+0.j 0.+0.j 1.+0.j 0.+0.j 0.+0.j 0.+0.j],
            [0.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j 1.+0.j 0.+0.j 0.+0.j],
            [0.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j 1.+0.j 0.+0.j],
            [0.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j 1.+0.j]
        ]
        >>> b = ComputationalBasis()
        >>> b.set_size(2)
        >>> b.pretty_print()
        Basis: [
            [1.+0.j 0.+0.j 0.+0.j 0.+0.j],
            [0.+0.j 1.+0.j 0.+0.j 0.+0.j],
            [0.+0.j 0.+0.j 1.+0.j 0.+0.j],
            [0.+0.j 0.+0.j 0.+0.j 1.+0.j]
        ]
    """

    def __init__(self, nb_qubits: Optional[int] = None):
        basis = []
        Basis.__init__(self, basis, nb_qubits)
        if nb_qubits is not None:
            self.set_size(nb_qubits)

    def set_size(self, nb_qubits: int):
        self.basis_vectors = [
            np.array([0] * i + [1] + [0] * (2**nb_qubits - 1 - i), dtype=np.complex64)
            for i in range(2**nb_qubits)
        ]
        self.nb_qubits = nb_qubits


class HadamardBasis(VariableSizeBasis):
    """Basis representing the Hadamard basis, also called X-basis or +/- basis.

    Args:
        nb_qubits: number of qubits in the basis

    Example:
        >>> HadamardBasis(2).pretty_print()
        Basis: [
            [0.5+0.j 0.5+0.j 0.5+0.j 0.5+0.j],
            [ 0.5+0.j -0.5+0.j  0.5+0.j -0.5+0.j],
            [ 0.5+0.j  0.5+0.j -0.5+0.j -0.5+0.j],
            [ 0.5+0.j -0.5+0.j -0.5+0.j  0.5-0.j]
        ]
    """

    def __init__(self, nb_qubits: Optional[int] = None):
        basis = []
        Basis.__init__(self, basis, nb_qubits)
        if nb_qubits is not None:
            self.set_size(nb_qubits)

    def set_size(self, nb_qubits: int):
        H = np.array([[1, 1], [1, -1]], dtype=np.complex64) / np.sqrt(2)
        Hn = reduce(np.kron, [H] * nb_qubits, np.eye(1))
        self.basis_vectors = [line for line in Hn]
        self.nb_qubits = nb_qubits



from __future__ import annotations
import re
from typing import Iterator, Sequence, TypeVar, Union, Callable, Iterable
import numpy as np
import numpy.typing as npt
from typeguard import typechecked

T = TypeVar("T")
ListOrSingle = Union[list[T], T]
"""Type alias for both elements of type ``T``, or list of elements of type ``T``."""
ArbitraryNestedSequence = Union[Sequence["ArbitraryNestedSequence"], T]
"""Since arbitrarily nested list are defined by recursion, this type allow us to 
define a base case.

Examples:
    >>> l = [[0,1],0,[[0,2],3]]
    >>> l = 1
    >>> l = [2,1,3]
"""
Matrix = Union[npt.NDArray[np.complex64], npt.NDArray[np.object_]]
"""Type alias denoting all the matrices we consider (either matrices of complex 
or of ``sympy`` expressions, given to ``numpy`` as objects)"""


@typechecked
def flatten_generator(lst: ArbitraryNestedSequence[T]) -> Iterator[T]:
    """Helper generator function for flattening an arbitrarily nested list.

    Args:
        lst: The list, or nested list, to be flattened.

    Yields:
        Elements from the input list in a flattened order.
    """
    if isinstance(lst, Sequence):
        for el in lst:
            yield from flatten_generator(el)  # type: ignore
    else:
        yield lst


@typechecked
def flatten(lst: ArbitraryNestedSequence[T]) -> list[T]:
    """Flattens an arbitrarily nested list.

    Example:
        >>> nested_list = [[1, 2, [3, 4]], [5, [6, 7]], 8]
        >>> flatten(nested_list)
        [1, 2, 3, 4, 5, 6, 7, 8]

    Args:
        lst: The list, or nested list, to be flattened.

    Returns:
        A flattened list containing all elements from the input list.
    """
    return list(flatten_generator(lst))


def one_lined_repr(obj: object):
    """One-liner returning a representation of the given object by removing
    extra whitespace.

    Args:
        obj: The object for which a representation is desired.
    """
    return re.sub(r"\s+", " ", repr(obj))


@typechecked
def find(iterable: Iterable[T], oracle: Callable[[T], bool]) -> T:
    """
    Finds the first element in the iterable that satisfies the given oracle.

    Example:
        >>> numbers = [1, 2, 3, 4, 5]
        >>> is_even = lambda x: x % 2 == 0
        >>> find(numbers, is_even)
        2

    Args:
        iterable: The iterable to search for the element.
        oracle: A callable function that takes an element and returns ``True``
            if the element satisfies the condition.

    Returns:
        The first element in the iterable that satisfies the oracle.

    Raises:
        ValueError: If no element in the iterable satisfies the given oracle.
    """
    for item in iterable:
        if oracle(item):
            return item
    raise ValueError("No objects satisfies the given oracle")



import numpy as np
import numpy.typing as npt
import pytest

from mpqp.core.instruction.gates.custom_gate import CustomGate
from mpqp.core.instruction.gates.gate_definition import UnitaryMatrix
from mpqp.tools.maths import is_unitary
from mpqp.gates import symbols


def test_custom_gate_created():
    definition = UnitaryMatrix(np.array([[1, 0], [0, 1j]]))
    assert is_unitary(CustomGate(definition, [0, 1]).to_matrix())


@pytest.mark.parametrize(
    "matrix_1, matrix_2",
    [
        (np.array([[1, 0], [0, -1]]), np.array([[3, 0], [0, -3.0]]) / 3),
    ],
)
def test_unitary_matrix_is_equivalent(
    matrix_1: npt.NDArray[np.complex64], matrix_2: npt.NDArray[np.complex64]
):
    unitary_matrix_1 = UnitaryMatrix(matrix_1)
    unitary_matrix_2 = UnitaryMatrix(matrix_2)
    assert unitary_matrix_1.is_equivalent(unitary_matrix_2)


@pytest.mark.parametrize(
    "matrix",
    [np.array([[symbols("theta"), 0], [0, 1]])],
)
def test_unitary_matrix_inverse_failing(matrix: npt.NDArray[np.complex64]):
    with pytest.raises(
        ValueError,
        match="Cannot invert arbitrary gates using symbolic variables",
    ):
        with pytest.warns(
            UserWarning,
            match="Cannot ensure that a operator defined with variables is unitary.",
        ):
            UnitaryMatrix(matrix).inverse()


@pytest.mark.parametrize(
    "matrix",
    [np.array([[1, 0], [0, 1]])],
)
def test_unitary_matrix_inverse_(matrix: npt.NDArray[np.complex64]):
    unit = UnitaryMatrix(matrix)
    assert unit.is_equivalent(unit.inverse())



from __future__ import annotations
from typing import Optional, Sequence

import pytest
import numpy as np
from qiskit import QuantumCircuit
from typeguard import TypeCheckError

from mpqp import QCircuit, Instruction, Barrier, Language
from mpqp.gates import Gate, CNOT, X, Y, Z, CZ, SWAP, T, H, Rx, S, Ry, Rz
from mpqp.measures import BasisMeasure, ExpectationMeasure, Observable
from mpqp.tools.generics import ListOrSingle, one_lined_repr

# 3M-TODO: a lot of these tests use str to test circuit equivalence, it would be
# preferable to define a __eq__ in the QCircuit class, or at least an
# `equivalent` method


@pytest.mark.parametrize(
    "init_param, printed_result_filename",
    [
        (0, "empty"),
        ([], "empty"),
        ([CNOT(0, 1), CNOT(1, 2), CNOT(0, 1)], "cnots"),
        ([CNOT(0, 1), X(0), BasisMeasure([0, 1], shots=100)], "measure"),
        ([CNOT(1, 2), Barrier(), X(0)], "barrier"),
    ],
)
def test_init_right(
    init_param: int | Sequence[Instruction], printed_result_filename: str
):
    with open(
        f"tests/core/test_circuit/init-{printed_result_filename}.txt",
        "r",
        encoding="utf-8",
    ) as f:
        assert str(QCircuit(init_param)) == f.read()


@pytest.mark.parametrize(
    "init_param",
    [1.0, X(0), -1],
)
def test_init_wrong(init_param: int | Sequence[Instruction]):
    with pytest.raises(TypeCheckError):
        QCircuit(init_param)


@pytest.mark.parametrize(
    "init_circuit, added_gates, printed_result_filename",
    [
        (QCircuit(2), X(0), "X"),
        (
            QCircuit([X(0)], nb_qubits=2),
            [CNOT(0, 1), BasisMeasure([0, 1], shots=100)],
            "list",
        ),
    ],
)
def test_add(
    init_circuit: QCircuit,
    added_gates: ListOrSingle[Instruction],
    printed_result_filename: str,
):
    with open(
        f"tests/core/test_circuit/add-{printed_result_filename}.txt",
        "r",
        encoding="utf-8",
    ) as f:
        init_circuit.add(added_gates)
        assert str(init_circuit) == f.read()


@pytest.mark.parametrize(
    "instructions, result",
    [
        ([CNOT(0, 1), CNOT(1, 2), CNOT(0, 1), CNOT(2, 3)], 3),
        ([CNOT(0, 1), CNOT(1, 2), CNOT(0, 1)], 3),
        ([CNOT(0, 1), CNOT(1, 2), CNOT(0, 1), X(2)], 3),
        ([CNOT(0, 1), CNOT(1, 2), CNOT(0, 1), Barrier(), X(2)], 4),
        ([CNOT(0, 1), CNOT(1, 2), CNOT(2, 3), X(0), X(0)], 3),
    ],
)
def test_depth(instructions: list[Instruction], result: int):
    assert QCircuit(instructions).depth() == result


@pytest.mark.parametrize(
    "circuit, result",
    [
        (QCircuit([CNOT(0, 1), CNOT(1, 2)]), (3, 0)),
        (QCircuit(3, nb_cbits=2), (3, 2)),
    ],
)
def test_size(circuit: QCircuit, result: tuple[int, Optional[int]]):
    assert circuit.size() == result


@pytest.mark.parametrize(
    "circuit, result",
    [
        (QCircuit([CNOT(0, 1), CNOT(1, 2), X(1), CNOT(1, 2)]), 4),
    ],
)
def test_len(circuit: QCircuit, result: int):
    assert len(circuit) == result


@pytest.mark.parametrize(
    "inst_list_1, inst_list_2",
    [
        ([CNOT(0, 1), CNOT(1, 2)], [X(1), CNOT(1, 2)]),
    ],
)
def test_append(inst_list_1: list[Instruction], inst_list_2: list[Instruction]):
    circ = QCircuit(inst_list_1)
    circ.append(QCircuit(inst_list_2))
    assert str(circ) == str(QCircuit(inst_list_1 + inst_list_2))
    assert str(QCircuit(inst_list_1) + QCircuit(inst_list_2)) == str(
        QCircuit(inst_list_1 + inst_list_2)
    )


@pytest.mark.parametrize(
    "inst_list_1, inst_list_2, inst_list_result",
    [
        (
            [CNOT(0, 1), CNOT(1, 2)],
            [X(1), CNOT(1, 2)],
            [CNOT(0, 1), CNOT(1, 2), X(4), CNOT(4, 5)],
        ),
    ],
)
def test_tensor(
    inst_list_1: list[Instruction],
    inst_list_2: list[Instruction],
    inst_list_result: list[Instruction],
):
    assert str(QCircuit(inst_list_1).tensor(QCircuit(inst_list_2))) == str(
        QCircuit(inst_list_result)
    )
    assert str(QCircuit(inst_list_1) @ (QCircuit(inst_list_2))) == str(
        QCircuit(inst_list_result)
    )


@pytest.mark.parametrize(
    "circuit, filter, count",
    [
        (
            QCircuit(
                [X(0), Y(1), Z(2), CNOT(0, 1), SWAP(0, 1), CZ(1, 2), X(2), X(1), X(0)]
            ),
            (),
            9,
        ),
        (
            QCircuit(
                [X(0), Y(1), Z(2), CNOT(0, 1), SWAP(0, 1), CZ(1, 2), X(2), X(1), X(0)]
            ),
            (X,),
            4,
        ),
    ],
)
def test_count(circuit: QCircuit, filter: tuple[type[Gate]], count: int):
    assert circuit.count_gates(*filter) == count


@pytest.mark.parametrize(
    "circuit, result_repr",
    [
        (
            QCircuit(
                [
                    BasisMeasure([0, 1], shots=1000),
                    ExpectationMeasure(
                        [1], Observable(np.identity(2, dtype=np.complex64)), shots=1000
                    ),
                ]
            ),
            "[BasisMeasure([0, 1], shots=1000), ExpectationMeasure([1], "
            "Observable(array([[1.+0.j, 0.+0.j], [0.+0.j, 1.+0.j]], dtype=complex64)), shots=1000)]",
        )
    ],
)
def test_get_measurements(circuit: QCircuit, result_repr: str):
    assert one_lined_repr(circuit.get_measurements()) == result_repr


@pytest.mark.parametrize(
    "circuit, printed_result_filename",
    [
        (QCircuit([X(0), CNOT(0, 1), BasisMeasure([0, 1], shots=100)]), "all"),
    ],
)
def test_without_measurements(circuit: QCircuit, printed_result_filename: str):
    with open(
        f"tests/core/test_circuit/wo_meas-{printed_result_filename}.txt",
        "r",
        encoding="utf-8",
    ) as f:
        assert str(circuit.without_measurements()) == f.read()


@pytest.mark.parametrize(
    "circuit, args, result_type, result_repr",
    [
        (
            QCircuit([X(0), CNOT(0, 1)]),
            (),
            QuantumCircuit,
            (
                "[CircuitInstruction(operation=Instruction(name='x', num_qubits=1,"
                " num_clbits=0, params=[]), qubits=(Qubit(QuantumRegister(2, 'q'), 0),),"
                " clbits=()), CircuitInstruction(operation=Instruction(name='cx',"
                " num_qubits=2, num_clbits=0, params=[]), qubits=(Qubit(QuantumRegister(2,"
                " 'q'), 0), Qubit(QuantumRegister(2, 'q'), 1)), clbits=())]"
            ),
        ),
    ],
)
def test_to_other_language(
    circuit: QCircuit, args: tuple[Language], result_type: type, result_repr: str
):
    qiskit_circuit = circuit.to_other_language(*args)
    assert type(qiskit_circuit) == QuantumCircuit
    assert repr(qiskit_circuit.data) == result_repr


@pytest.mark.parametrize(
    "circuit, printed_result_filename",
    [(QCircuit([X(0), CNOT(0, 1), BasisMeasure([0, 1], shots=100)]), "all")],
)
def test_to_qasm_2(circuit: QCircuit, printed_result_filename: str):
    with open(
        f"tests/core/test_circuit/{printed_result_filename}.qasm2",
        "r",
        encoding="utf-8",
    ) as f:
        assert circuit.to_qasm2() == f.read()


@pytest.mark.parametrize(
    "circuit, printed_result_filename",
    [
        (QCircuit([X(0), CNOT(0, 1), BasisMeasure([0, 1], shots=100)]), "all"),
        (
            QCircuit(
                [
                    T(0),
                    CNOT(0, 1),
                    X(0),
                    H(1),
                    Z(2),
                    CZ(2, 1),
                    SWAP(2, 0),
                    CNOT(0, 2),
                    Ry(3.14 / 2, 2),
                    S(1),
                    H(3),
                    CNOT(1, 2),
                    Rx(3.14, 1),
                    CNOT(3, 0),
                    Rz(3.14, 0),
                    BasisMeasure([0, 1, 2, 3], shots=2000),
                ]
            ),
            "lot_of_gates",
        ),
    ],
)
def test_to_qasm_3(circuit: QCircuit, printed_result_filename: str):
    with open(
        f"tests/core/test_circuit/{printed_result_filename}.qasm3",
        "r",
        encoding="utf-8",
    ) as f:
        assert circuit.to_qasm3().strip() == f.read().strip()



from braket.circuits import Operator, Circuit
from braket.circuits.gates import H, CNot
import pytest

from mpqp.qasm.qasm_to_braket import qasm3_to_braket_Circuit


@pytest.mark.parametrize(
    "qasm_code, braket_operators",
    [
        (
            """OPENQASM 3.0;""",
            [],
        ),
        (
            """OPENQASM 3.0;
            include 'stdgates.inc';

            qubit[2] q;
            bit[2] c;
            h q[0];
            cx q[0],q[1];

            c[0] = measure q[0];
            c[1] = measure q[1];""",
            [H(), CNot()],
        ),
    ],
)
def test_qasm3_to_braket_Circuit(qasm_code: str, braket_operators: list[Operator]):
    circ = qasm3_to_braket_Circuit(qasm_code)
    assert isinstance(circ, Circuit)
    for circ_instr, expected_operator in zip(circ.instructions, braket_operators):
        assert circ_instr.operator == expected_operator



import pytest

import numpy as np
import numpy.typing as npt

from mpqp.measures import (
    Basis,
    ComputationalBasis,
    HadamardBasis,
    VariableSizeBasis,
)


@pytest.mark.parametrize(
    "init_vectors, size",
    [
        ([np.array([1, 0]), np.array([0, -1])], 1),
    ],
)
def test_right_init_basis(init_vectors: list[npt.NDArray[np.complex64]], size: int):
    b = Basis(init_vectors)
    assert b.nb_qubits == size
    assert (b.basis_vectors[i] == init_vectors[i] for i in range(len(init_vectors)))


@pytest.mark.parametrize(
    "init_vectors, part_of_error",
    [
        ([np.array([1, 0]), np.array([0, -1]), np.array([0, -1])], "number of vector"),
        ([np.array([1, 0])], "number of vector"),
        ([np.array([1, 0, 0]), np.array([0, -1])], "same size"),
        ([np.array([1, 1]), np.array([1, -1])], "normalized"),
        ([np.array([1, 1]) / np.sqrt(2), np.array([0, -1])], "orthogonal"),
    ],
)
def test_wrong_init_basis(
    init_vectors: list[npt.NDArray[np.complex64]], part_of_error: str
):
    with pytest.raises(ValueError) as error:
        Basis(init_vectors)
    assert part_of_error in error.exconly()


@pytest.mark.parametrize(
    "basis, size, is_initialized, result_pp",
    [
        (
            ComputationalBasis,
            3,
            True,
            (
                "Basis: [\n"
                "    [1.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j],\n"
                "    [0.+0.j 1.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j],\n"
                "    [0.+0.j 0.+0.j 1.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j],\n"
                "    [0.+0.j 0.+0.j 0.+0.j 1.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j],\n"
                "    [0.+0.j 0.+0.j 0.+0.j 0.+0.j 1.+0.j 0.+0.j 0.+0.j 0.+0.j],\n"
                "    [0.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j 1.+0.j 0.+0.j 0.+0.j],\n"
                "    [0.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j 1.+0.j 0.+0.j],\n"
                "    [0.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j 1.+0.j]\n"
                "]\n"
            ),
        ),
        (
            ComputationalBasis,
            2,
            False,
            (
                "Basis: [\n"
                "    [1.+0.j 0.+0.j 0.+0.j 0.+0.j],\n"
                "    [0.+0.j 1.+0.j 0.+0.j 0.+0.j],\n"
                "    [0.+0.j 0.+0.j 1.+0.j 0.+0.j],\n"
                "    [0.+0.j 0.+0.j 0.+0.j 1.+0.j]\n"
                "]\n"
            ),
        ),
        (
            HadamardBasis,
            2,
            True,
            (
                "Basis: [\n"
                "    [0.5+0.j 0.5+0.j 0.5+0.j 0.5+0.j],\n"
                "    [ 0.5+0.j -0.5+0.j  0.5+0.j -0.5+0.j],\n"
                "    [ 0.5+0.j  0.5+0.j -0.5+0.j -0.5+0.j],\n"
                "    [ 0.5+0.j -0.5+0.j -0.5+0.j  0.5-0.j]\n"
                "]\n"
            ),
        ),
    ],
)
def test_basis_implementations(
    basis: type[VariableSizeBasis],
    size: int,
    is_initialized: bool,
    result_pp: str,
    capsys: pytest.CaptureFixture[str],
):
    if is_initialized:
        b = basis(size)
    else:
        b = basis()
        b.set_size(size)
    b.pretty_print()
    captured = capsys.readouterr()
    assert captured.out == result_pp



import pytest

from mpqp.core.instruction.measurement.basis import ComputationalBasis, HadamardBasis
from mpqp.core.instruction.measurement.basis_measure import BasisMeasure
from mpqp.core.languages import Language


def test_basis_measure_init():
    measure = BasisMeasure([0, 1], shots=1024, basis=ComputationalBasis())
    assert measure.targets == [0, 1]
    assert measure.shots == 1024
    assert isinstance(measure.basis, ComputationalBasis)


def test_basis_measure_init_fails_duplicate_c_targets():
    with pytest.raises(ValueError, match="Duplicate registers in targets"):
        BasisMeasure(targets=[0, 1], c_targets=[2, 2, 3], shots=1024)


def test_basis_measure_to_other_language_not_implemented():
    measure = BasisMeasure([0], basis=HadamardBasis)
    with pytest.raises(NotImplementedError):
        measure.to_other_language(language=Language.QISKIT)


def test_basis_measure_repr():
    measure = BasisMeasure([0, 1], shots=1024)
    representation = repr(measure)
    assert representation == "BasisMeasure([0, 1], shots=1024)"



from typing import Any
import numpy as np
import pytest
from sympy import Expr

from mpqp import QCircuit
from mpqp.core.instruction.measurement.expectation_value import (
    ExpectationMeasure,
    Observable,
)
from mpqp.gates import *
from mpqp.execution.devices import AWSDevice, AvailableDevice, IBMDevice, ATOSDevice
from mpqp.execution.vqa import minimize, Optimizer
from mpqp.execution.vqa.vqa import OptimizableFunc
from mpqp.execution.runner import _run_single  # pyright: ignore[reportPrivateUsage]

# the symbols function is a bit wacky, so some manual type definition is needed here
theta: Expr = symbols("θ")  # type: ignore


def with_local_devices(args: tuple[Any, ...]):
    return (
        (*args, d)
        for d in list(IBMDevice) + list(ATOSDevice) + list(AWSDevice)
        if not d.is_remote() and d.is_gate_based()
    )


@pytest.mark.parametrize(
    "circ, minimum, device",
    with_local_devices(
        (
            QCircuit(
                [
                    P(theta, 0),
                    ExpectationMeasure([0], Observable(np.array([[0, 1], [1, 0]]))),
                ]
            ),
            0,
        )
    ),
)
def test_optimizer_circuit(circ: QCircuit, minimum: float, device: AvailableDevice):
    try:
        assert minimize(circ, Optimizer.BFGS, device)[0] - minimum < 0.05
    except (ValueError, NotImplementedError) as err:
        if "not handled" not in str(err):
            raise


@pytest.mark.parametrize(
    "eval_f, minimum, device",
    with_local_devices(
        (
            lambda params: (
                1
                - _run_single(
                    QCircuit(
                        [
                            P(theta, 0),
                            ExpectationMeasure(
                                [0], Observable(np.array([[0, 1], [1, 0]]))
                            ),
                        ]
                    ),
                    ATOSDevice.MYQLM_PYLINALG,
                    {theta: params[0]},
                ).expectation_value
                ** 2
            ),
            1,
        )
    ),
)
def test_optimizer_func(
    eval_f: OptimizableFunc, minimum: float, device: AvailableDevice
):
    try:
        assert minimize(eval_f, Optimizer.BFGS, nb_params=1)[0] - minimum < 0.05
    except (ValueError, NotImplementedError) as err:
        if "not handled" not in str(err):
            raise



"""Demonstration MPQP"""

import numpy as np
from mpqp.gates import *
from mpqp import QCircuit
from mpqp.measures import BasisMeasure
from mpqp.execution import run
from mpqp.execution.devices import ATOSDevice, IBMDevice, AWSDevice
from mpqp.tools.visualization import plot_results_sample_mode
import matplotlib.pyplot as plt


# Constructing the circuit
meas = BasisMeasure(list(range(3)), shots=2000)
c = QCircuit([T(0), CNOT(0, 1), Ry(np.pi / 2, 2), S(1), CZ(2, 1), SWAP(2, 0), meas])
print(c)

# Run the circuit on a selected device
results = run(
    c,
    [
        IBMDevice.AER_SIMULATOR,
        ATOSDevice.MYQLM_PYLINALG,
        ATOSDevice.MYQLM_CLINALG,
        AWSDevice.BRAKET_LOCAL_SIMULATOR,
    ],
)

# Print the results
print(results)

# Display the circuit
plot_results_sample_mode(results)
plt.show()

c = QCircuit([T(0), CNOT(0, 1), Ry(np.pi / 2, 2), S(1), CZ(2, 1), SWAP(2, 0)])
res = run(
    c,
    [
        IBMDevice.AER_SIMULATOR_STATEVECTOR,
        ATOSDevice.MYQLM_PYLINALG,
        ATOSDevice.MYQLM_CLINALG,
        AWSDevice.BRAKET_LOCAL_SIMULATOR,
    ],
)
# Print the results
print(res)



"""Collection of functions to plot, with matplotlib, one or several Results as 
histograms, when the ``job_type`` is ``SAMPLE``."""

import math
import random
from typing import Union

import matplotlib.pyplot as plt
from typeguard import typechecked

from mpqp.execution.job import JobType
from mpqp.execution.result import Result, BatchResult


@typechecked
def plot_results_sample_mode(results: Union[BatchResult, Result]):
    """
    Display the several results in parameter in a grid using subplots.

    Args:
        results: BatchResult to plot in the same window.
    """
    if isinstance(results, Result):
        return _plot_result_sample_mode(results)

    n_cols = math.ceil(len(results.results) // 2)
    n_rows = 2

    for index, result in enumerate(results.results):
        plt.subplot(n_rows, n_cols, index + 1)

        _prep_plot(result)

    plt.tight_layout()


@typechecked
def _plot_result_sample_mode(result: Result):
    """
    Display the result in parameter in a single figure.

    Args:
        result: Result to plot.
    """
    plt.figure()

    _prep_plot(result)


@typechecked
def _prep_plot(result: Result):
    """
    Extract sampling info from the result and construct the bar diagram plot.

    Args:
        result: Result to transform into a bar diagram.
    """

    x_array, y_array = _result_to_array_sample_mode(result)
    x_axis = range(len(x_array))

    plt.bar(x_axis, y_array, color=(*[random.random() for _ in range(3)], 0.9))
    plt.xticks(x_axis, x_array, rotation=-60)
    plt.xlabel("State")
    plt.ylabel("Counts")
    device = result.job.device
    plt.title(type(device).__name__ + ", " + device.name)


@typechecked
def _result_to_array_sample_mode(result: Result):
    """
    Transform a result into an x and y array containing the string of basis state with the associated counts
    Args:
        result: result used to generate the array

    Returns:
        the tuple x_array (strings for each basis state) and y_array (counts for each basis state)
    """
    if result.job.job_type != JobType.SAMPLE:
        raise NotImplementedError(
            f"{result.job.job_type} not handled, only {JobType.SAMPLE} is handled for now."
        )
    if result.job.measure is None:
        raise ValueError(
            f"{result.job=} has no measure, making the counting impossible"
        )
    n = result.job.measure.nb_qubits
    x_array = [f"|{bin(i)[2:].zfill(n)}⟩" for i in range(2**n)]
    y_array = result.counts
    return x_array, y_array



from mpqp.gates import CNOT, H, Rx, Ry, Rz
from mpqp import QCircuit
from mpqp.execution import run
from mpqp.execution.devices import ATOSDevice
import numpy as np

circuit = QCircuit(5)

circuit.add(H(0))
circuit.add(H(1))
circuit.add(Rz(np.pi / 3, 1))
circuit.add(H(2))
circuit.add(Rx(np.pi / 3, 2))
circuit.add(H(3))
circuit.add(H(4))
circuit.add(CNOT(0, 1))
circuit.add(CNOT(1, 2))
circuit.add(CNOT(2, 3))
circuit.add(CNOT(3, 4))
circuit.add(Ry(np.pi / 3, 4))

print(run(circuit, ATOSDevice.MYQLM_PYLINALG).state_vector)


circuit2 = QCircuit(5)

circuit2.add(H(0))
circuit2.add(H(1))
circuit2.add(Rz(np.pi / 3, 1))
circuit2.add(H(2))
circuit2.add(Rx(np.pi / 3, 2))
circuit2.add(H(3))
circuit2.add(H(4))
circuit2.add(CNOT(0, 1))
circuit2.add(CNOT(2, 3))
circuit2.add(CNOT(1, 2))
circuit2.add(CNOT(3, 4))
circuit2.add(Ry(np.pi / 3, 4))

print(run(circuit2, ATOSDevice.MYQLM_PYLINALG).state_vector)





import warnings
import numpy as np
import pytest

from mpqp.measures import Observable, ExpectationMeasure


@pytest.mark.parametrize(
    "targets",
    [[0, 1, 2], [1, 2, 3], list(range(4, 10))],
)
def test_expectation_measure_right_targets(targets: list[int]):
    obs = Observable(np.diag([1] * 2 ** len(targets)))
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        ExpectationMeasure(targets, obs)


@pytest.mark.parametrize(
    "targets, expected_swaps",
    [
        ([1, 3, 4], [{2, 3}, {3, 4}]),
        ([1, 0, 2], [{1, 0}]),
        ([2, 0, 3], [{0, 2}, {1, 2}, {2, 3}]),
    ],
)
def test_expectation_measure_wrong_targets(
    targets: list[int], expected_swaps: list[tuple[int, int]]
):
    obs = Observable(np.diag([1] * 2 ** len(targets)))
    with pytest.warns(UserWarning):
        measure = ExpectationMeasure(targets, obs)
    assert [
        set(swap.targets) for swap in measure.pre_measure.instructions
    ] == expected_swaps



from __future__ import annotations
from abc import abstractmethod, ABC
from copy import deepcopy
from typing import Optional

import numpy as np
from scipy.linalg import fractional_matrix_power
from typeguard import typechecked

from mpqp.core.instruction.instruction import Instruction
from mpqp.core.instruction.gates.gate_definition import UnitaryMatrix
from mpqp.tools.generics import Matrix
from mpqp.tools.maths import matrix_eq


@typechecked
class Gate(Instruction, ABC):
    """Represent a unitary operator acting on qubit(s).

    A gate is an measurement and the main component of a circuit. The semantics
    of a gate is defined using
    :class:`GateDefinition<mpqp.core.instruction.gates.gate_definition.GateDefinition>`.

    Args:
        targets: List of indices referring to the qubits on which the gate will
            be applied.
        label: Label used to identify the gate.
    """

    def __init__(
        self,
        targets: list[int],
        label: Optional[str] = None,
    ):
        super().__init__(targets, label=label)

    @abstractmethod
    def to_matrix(self) -> Matrix:
        """Return the "base" matricial semantics to this gate. Without
        considering potential column and row permutations needed if the targets
        of the gate are not sorted.

        Example:
            >>> gd = UnitaryMatrix(
            ...     np.array([[0, 0, 0, 1], [0, 1, 0, 0], [1, 0, 0, 0], [0, 0, 1, 0]])
            ... )
            >>> CustomGate(gd, [1, 2]).to_matrix()
            array([[0, 0, 0, 1],
                   [0, 1, 0, 0],
                   [1, 0, 0, 0],
                   [0, 0, 1, 0]])
            >>> SWAP(0,1).to_matrix()
            array([[1, 0, 0, 0],
                   [0, 0, 1, 0],
                   [0, 1, 0, 0],
                   [0, 0, 0, 1]])

        Returns:
            A numpy array representing the unitary matrix of the gate.
        """

    def inverse(self) -> Gate:
        """Computing the inverse of this gate.

        Example:
            >>> Z(0).inverse()
            Z(0)
            >>> CustomGate(UnitaryMatrix(np.diag([1,1j])),[0]).inverse().to_matrix()
            array([[1,  0 ],
                   [0, -1j]])

        Returns:
            The gate corresponding to the inverse of this gate.
        """
        from mpqp.core.instruction.gates.custom_gate import CustomGate

        return CustomGate(
            UnitaryMatrix(self.to_matrix().transpose().conjugate()),
            self.targets,
            self.label,
        )

    def is_equivalent(self, other: Gate) -> bool:
        """Determine if the gate in parameter is equivalent to this gate.

        The equivalence of two gate is only determined from their matricial
        semantics (and thus ignores all other aspects of the gate such as the
        target qubits, the label, etc....)

        Example:
            >>> X(0).is_equivalent(CustomGate(UnitaryMatrix(np.array([[0,1],[1,0]])),[1]))
            True

        Args:
            other: the gate to test if it is equivalent to this gate

        Returns:
            ``True`` if the two gates' matrix semantics are equal.
        """
        return matrix_eq(self.to_matrix(), other.to_matrix())

    def power(self, exponent: float) -> Gate:
        """Compute the exponentiation `G^{exponent}` of this gate G.

        Examples:
            >>> swap_gate = SWAP(0,1)
            >>> (swap_gate.power(2)).to_matrix()
            array([[1, 0, 0, 0],
                   [0, 1, 0, 0],
                   [0, 0, 1, 0],
                   [0, 0, 0, 1]])
            >>> (swap_gate.power(-1)).to_matrix()
            array([[1, 0, 0, 0],
                   [0, 0, 1, 0],
                   [0, 1, 0, 0],
                   [0, 0, 0, 1]])
            >>> (swap_gate.power(0.75)).to_matrix() # not implemented yet
            array([[1.        +0.j        , 0.        +0.j        ,
                    0.        +0.j        , 0.        +0.j        ],
                   [0.        +0.j        , 0.14644661+0.35355339j,
                    0.85355339-0.35355339j, 0.        +0.j        ],
                   [0.        +0.j        , 0.85355339-0.35355339j,
                    0.14644661+0.35355339j, 0.        +0.j        ],
                   [0.        +0.j        , 0.        +0.j        ,
                    0.        +0.j        , 1.        +0.j        ]])

        Args:
            exponent: Number representing the exponent.

        Returns:
            The gate elevated to the exponent in parameter.
        """
        from mpqp.core.instruction.gates.custom_gate import CustomGate

        return CustomGate(
            definition=UnitaryMatrix(
                fractional_matrix_power(self.to_matrix(), exponent)
            ),
            targets=self.targets,
            label=None if self.label is None else self.label + f"^{exponent}",
        )

    def tensor_product(self, other: Gate) -> Gate:
        """Compute the tensor product of the current gate.

        Example:
            >>> (X(0).tensor_product(Z(0))).to_matrix()
            array([[ 0,  0,  1,  0],
                   [ 0,  0,  0, -1],
                   [ 1,  0,  0,  0],
                   [ 0, -1,  0,  0]])

        Args:
            other: Second operand of the tensor product.

        Returns:
            A Gate representing a tensor product of this gate with the gate in
            parameter.

        # 3M-TODO: to be implemented, don't trust the code bellow, it's pure experiments
        """
        from mpqp.core.instruction.gates.custom_gate import CustomGate

        gd = UnitaryMatrix(
            np.kron(self.to_matrix(), other.to_matrix())
        )  # self, gate, type="tensor"

        # compute the list of qubits that will be targeted by these gates
        ...

        # instantiate the definition

        l1 = "g1" if self.label is None else self.label
        l2 = "g2" if self.label is None else self.label

        return CustomGate(
            definition=gd,
            targets=[0],
            label=f"{l1}⊗{l2}",
        )

    def _mandatory_label(self, postfix: str = ""):
        return "g" + postfix if self.label is None else self.label

    def product(self, other: Gate, targets: Optional[list[int]] = None) -> Gate:
        """Compute the composition of self and the other gate.

        Example:
            >>> (X(0).product(Z(0))).to_matrix()
            array([[ 0, -1],
                   [ 1,  0]])

        Args:
            other: Rhs of the product.
            targets: Qubits on which this new gate will operate. If not given,
                the targets of the two gates multiplied must be the same and the
                resulting gate will have this same targets.

        Returns:
            The product of the two gates concerned.
        """
        from mpqp.core.instruction.gates.custom_gate import CustomGate

        return CustomGate(
            definition=UnitaryMatrix(self.to_matrix().dot(other.to_matrix())),  # type: ignore
            targets=self._check_targets_compatibility(other, targets),
            label=f"{self._mandatory_label('1')}×{other._mandatory_label('2')}",
        )

    def scalar_product(self, scalar: complex) -> Gate:
        """Multiply this gate by a scalar. It normalizes the subtraction
        to ensure it is unitary.

        Example:
            >>> (X(0).scalar_product(1j)).to_matrix()
            array([[0, 1j],
                   [1j, 0]])

        Args:
            scalar: The number to multiply the gate's matrix by.

        Returns:
            The result of the multiplication, the targets of the resulting gate
            will be the same as the ones of the initial gate.
        """
        from mpqp.core.instruction.gates.custom_gate import CustomGate

        return CustomGate(
            UnitaryMatrix(self.to_matrix() * scalar / abs(scalar)),
            targets=self.targets,
            label=f"{scalar}×{self._mandatory_label()}",
        )

    def minus(self, other: Gate, targets: Optional[list[int]] = None) -> Gate:
        """Compute the subtraction of two gates. It normalizes the subtraction
        to ensure it is unitary.

        Example:
            >>> (X(0).minus(Z(0))).to_matrix()
            array([[-0.70710678,  0.70710678],
                   [ 0.70710678,  0.70710678]])

        Args:
            other: The gate to subtract to this gate.
            targets: Qubits on which this new gate will operate. If not given,
                the targets of the two gates multiplied must be the same and the
                resulting gate will have this same targets.

        Returns:
            The subtraction of ``self`` and ``other``.

        """
        from mpqp.core.instruction.gates.custom_gate import CustomGate

        subtraction = self.to_matrix() - other.to_matrix()
        return CustomGate(
            definition=UnitaryMatrix(subtraction / np.linalg.norm(subtraction, 2)),  # type: ignore
            targets=self._check_targets_compatibility(other, targets),
            label=f"{self._mandatory_label('1')}-{other._mandatory_label('2')}",
        )

    def plus(self, other: Gate, targets: Optional[list[int]] = None) -> Gate:
        """Compute the sum of two gates. It normalizes the subtraction
        to ensure it is unitary.

        Example:
            >>> (X(0).plus(Z(0))).to_matrix()
            array([[ 0.70710678,  0.70710678],
                   [ 0.70710678, -0.70710678]])

        Args:
            other: The gate to add to this gate.
            targets: Qubits on which this new gate will operate. If not given, the targets of the two gates multiplied
            must be the same and the resulting gate will have this same targets.

        Returns:
            The sum of ``self`` and ``other``.
        """
        from mpqp.core.instruction.gates.custom_gate import CustomGate

        addition = self.to_matrix() + other.to_matrix()
        return CustomGate(
            definition=UnitaryMatrix(addition / np.linalg.norm(addition, 2)),  # type: ignore
            targets=self._check_targets_compatibility(other, targets),
            label=f"{self._mandatory_label('1')}+{other._mandatory_label('2')}",
        )

    def _check_targets_compatibility(self, other: Gate, targets: Optional[list[int]]):
        if targets is None:
            if self.targets != other.targets:
                raise ValueError(
                    "Cannot infer what targets to use, please specify them in the "
                    "`targets` argument."
                )
            targets = self.targets
        if self.nb_qubits != other.nb_qubits:
            raise ValueError(
                f"Incompatible shapes for gates: respectively {self.nb_qubits} "
                f"and {other.nb_qubits} qubits"
            )
        if len(targets) != self.nb_qubits:
            raise ValueError(
                f"Incorrect size for targets: size {len(targets)} while it "
                f"should be {self.nb_qubits}"
            )
        return targets

    def __add__(self, other: Gate) -> Gate:
        return self.plus(other)

    def __sub__(self, other: Gate) -> Gate:
        return self.minus(other)


@typechecked
class InvolutionGate(Gate, ABC):
    """Gate who's inverse is itself.

    Args:
        targets: List of indices referring to the qubits on which the gate will be applied.
        label: Label used to identify the gate.
    """

    def inverse(self) -> Gate:
        return deepcopy(self)


@typechecked
class SingleQubitGate(Gate, ABC):
    """Gates operating on a single qubit.

    Args:
        target: Index or referring to the qubit on which the gate will be applied.
        label: Label used to identify the gate.
    """

    def __init__(self, target: int, label: Optional[str] = None):
        Gate.__init__(self, [target], label)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.targets[0]})"



# pyright: reportUnusedImport=false
from .basis import ComputationalBasis, Basis, HadamardBasis, VariableSizeBasis
from .basis_measure import BasisMeasure
from .expectation_value import ExpectationMeasure, Observable
from .measure import Measure



import math
from typing import Optional
from typeguard import typechecked

import numpy as np
from braket.circuits.observables import Hermitian
from braket.tasks import GateModelQuantumTaskResult, QuantumTask
from braket.aws import AwsQuantumTask

from mpqp import QCircuit, Language
from mpqp.core.instruction.measurement import ExpectationMeasure, BasisMeasure, Observable
from mpqp.execution.connection.aws_connection import get_braket_device
from mpqp.execution.devices import AWSDevice
from mpqp.execution.result import Result, Sample, StateVector
from mpqp.tools.errors import AWSBraketRemoteExecutionError
from mpqp.execution.job import Job, JobType, JobStatus
from mpqp.qasm.qasm_to_braket import qasm3_to_braket_Circuit
from mpqp.tools.errors import DeviceJobIncompatibleError


@typechecked
def run_braket(job: Job) -> Result:
    """
    Executes the job on the right AWS Braket device (local or remote) precised in the job in parameter and waits until
    the task is completed, then returns the Result.
    This function is not meant to be used directly, please use ``runner.run(...)`` instead.

    Args:
        job: Job to be executed.

    Returns:
        A Result after submission and execution of the job.
    """
    _, task = submit_job_braket(job)
    assert isinstance(job.device, AWSDevice)
    return extract_result(task.result(), job, job.device)


@typechecked
def submit_job_braket(job: Job) -> tuple[str, QuantumTask]:
    """
    Submits the job to the right local/remote device and returns the generated task.
    This function is not meant to be used directly, please use ``runner.submit(...)`` instead.

    Args:
        job: Job to be executed.

    Returns:
        A string representing the task' id, and the Task itself.
    """

    # check some compatibility issues

    if job.job_type == JobType.STATE_VECTOR and job.device.is_remote():
        raise DeviceJobIncompatibleError("State vector cannot be computed using AWS Braket remote simulators and "
                                         "devices. Please use the LocalSimulator instead")

    # instantiate the device
    device = get_braket_device(job.device)  # type: ignore

    # convert job circuit into braket circuit
    brkt_circuit = job.circuit.to_other_language(Language.BRAKET)

    if job.job_type == JobType.STATE_VECTOR:
        brkt_circuit.state_vector()  # type: ignore
        job.status = JobStatus.RUNNING
        task = device.run(brkt_circuit, shots=0)

    elif job.job_type == JobType.SAMPLE:
        job.status = JobStatus.RUNNING
        task = device.run(brkt_circuit, shots=job.measure.shots)

    elif job.job_type == JobType.OBSERVABLE:
        if not isinstance(job.measure, ExpectationMeasure):
            raise ValueError(
                "Cannot compute expectation value if measure used in job is not of "
                "type ExpectationMeasure"
            )

        herm_op = Hermitian(job.measure.observable.matrix)
        brkt_circuit.expectation(observable=herm_op, target=job.measure.targets)  # type: ignore

        job.status = JobStatus.RUNNING
        task = device.run(brkt_circuit, shots=job.measure.shots)

    else:
        raise NotImplementedError(f"Job of type {job.job_type} not handled.")

    return task.id, task


@typechecked
def extract_result(braket_result: GateModelQuantumTaskResult, job: Optional[Job] = None,
                   device: Optional[AWSDevice] = AWSDevice.BRAKET_LOCAL_SIMULATOR) -> Result:
    """
    Constructs a Result from the result given by the run with Braket.

    Args:
        braket_result: Result returned by myQLM/QLM after running of the job.
        job: Original mpqp job used to generate the run. Used to retrieve more easily info to instantiate the result.
        device: AWSDevice on which the job was submitted.

    Returns:
        A Result containing the result info extracted from the Braket result.
    """
    if job is None:
        if len(braket_result.values) == 0:
            job_type = JobType.SAMPLE
            nb_qubits = len(list(braket_result.measurement_counts.keys())[0])
            shots = braket_result.task_metadata.shots
            measure = BasisMeasure(list(range(nb_qubits)), shots=shots)
        elif isinstance(braket_result.values[0], float):
            job_type = JobType.OBSERVABLE
            nb_qubits = braket_result.task_metadata.deviceParameters.paradigmParameters.qubitCount
            shots = braket_result.task_metadata.shots
            measure = ExpectationMeasure(list(range(nb_qubits)),
                                         Observable(np.zeros((2 ** nb_qubits, 2 ** nb_qubits))),
                                         shots)
        else:
            job_type = JobType.STATE_VECTOR
            nb_qubits = int(math.log2(len(braket_result.values[0])))
            measure = BasisMeasure(list(range(nb_qubits)), shots=0)
        job = Job(job_type, QCircuit(nb_qubits), device, measure)
    job.status = JobStatus.DONE

    if job.job_type == JobType.STATE_VECTOR:
        vector = braket_result.values[0]
        state_vector = StateVector(vector, nb_qubits=job.circuit.nb_qubits)
        return Result(job, state_vector, 0, 0)

    elif job.job_type == JobType.SAMPLE:
        counts = braket_result.measurement_counts
        sample_info = []
        for state in counts.keys():
            sample_info.append(
                Sample(job.circuit.nb_qubits, count=counts[state], bin_str=state)
            )
        return Result(job, sample_info, None, job.measure.shots)

    elif job.job_type == JobType.OBSERVABLE:
        exp_value = braket_result.values[0]
        return Result(job, exp_value, None, job.measure.shots)

    else:
        raise NotImplementedError(f"Job of type {job.job_type} not handled.")


@typechecked
def get_result_from_aws_task_arn(task_arn: str = None) -> Result:
    """
    Retrieves the result, described by the job_id in parameter, from the remote QLM and converts it into an mpqp result.
    If the job is still running, we wait (blocking) until it is DONE.

    Args:
        task_arn: Arn of the remote aws task.

    """
    task: QuantumTask = AwsQuantumTask(task_arn)
    # eventually catch an error if the id is not correct (wrong ID, wrong region/blablabla)

    status: str = task.state()  # get the status of the task, it is a string
    # depending on the status, either raise an error, wait for the task to finish, or get the result if done

    if status in ["FAILED", "CANCELLED"]:
        raise AWSBraketRemoteExecutionError("")
    elif status in ["CREATED", "QUEUED", "RUNNING", "COMPLETED"]:  #
        result = task.result()  # will get the result, and eventually wait for it
    else:
        raise AWSBraketRemoteExecutionError(
            f"Unknown status {status} for the task {task_arn}"
        )

    device_arn = task.metadata()["deviceArn"]
    device = AWSDevice.from_arn(device_arn)

    return extract_result(result, None, device)



from __future__ import annotations

from abc import ABC
from typing import Optional

from typeguard import typechecked

from .gate import Gate


@typechecked
class ControlledGate(Gate, ABC):
    """Abstract class representing a controlled gate, that can be controlled by
    one or several qubits.

    Args:
        controls: List of indices referring to the qubits used to control the gate.
        targets: List of indices referring to the qubits on which the gate will be applied.
        non_controlled_gate: The original, non controlled, gate.
        label: Label used to identify the gate.
    """

    def __init__(
        self,
        controls: list[int],
        targets: list[int],
        non_controlled_gate: Optional[Gate] = None,
        label: Optional[str] = None,
    ):
        if len(set(controls)) != len(controls):
            raise ValueError(f"Duplicate registers in controls: {controls}")
        if len(set(controls).intersection(set(targets))):
            raise ValueError(
                f"Common registers between targets {targets} and controls {controls}"
            )
        self.controls = controls
        """See parameter description."""
        self.non_controlled_gate = non_controlled_gate
        """See parameter description."""

        Gate.__init__(self, targets, label)



"""Example 1: Building a Bell state"""

from mpqp.gates import H, CNOT
from mpqp import QCircuit
from mpqp.measures import BasisMeasure
from mpqp.execution.devices import IBMDevice, AWSDevice
from mpqp.execution import run
from mpqp.tools.visualization import plot_results_sample_mode
import matplotlib.pyplot as plt

# Declaration of the circuit with the right size
circuit = QCircuit(2, label="Bell pair")
# Constructing the circuit by adding gates and measurements
circuit.add(H(0))
circuit.add(CNOT(0, 1))
circuit.add(BasisMeasure([0, 1], shots=1000))
results = run(circuit, [IBMDevice.AER_SIMULATOR, AWSDevice.BRAKET_LOCAL_SIMULATOR])
print(results)

plot_results_sample_mode(results)
circuit.display()
plt.show()
